"""
Test various functions regarding chapter 4: Sampling (Bootstrapping, Concurrency).
"""

import unittest

import numpy as np
import pandas as pd

from famlafl.sampling.bootstrapping import seq_bootstrap, get_ind_matrix, get_ind_mat_average_uniqueness, \
    _bootstrap_loop_run, get_ind_mat_label_uniqueness  # pylint: disable=protected-access
from famlafl.sampling.concurrent import get_av_uniqueness_from_triple_barrier, num_concurrent_events


def book_ind_mat_implementation(bar_index, label_endtime):
    """
    Book implementation of get_ind_matrix function
    """
    ind_mat = pd.DataFrame(0, index=bar_index, columns=range(label_endtime.shape[0]))
    for i, (start, end) in enumerate(label_endtime.iteritems()):
        ind_mat.loc[start:end, i] = 1.
    return ind_mat


class TestSampling(unittest.TestCase):
    """
    Test Triple barrier, meta-labeling, dropping rare labels, and daily volatility.
    """

    def setUp(self):
        """
        Set samples_info_sets (t1), price bars
        """
        self.price_bars = pd.Series(index=pd.date_range(start="1/1/2018", end='1/8/2018', freq='H'), dtype='float64')
        self.samples_info_sets = pd.DataFrame(index=self.price_bars.index[[1, 2, 5, 7, 10, 11, 12, 20]])
        self.samples_info_sets['t1'] = self.samples_info_sets.index + pd.Timedelta('2H')

    def test_num_concurrent_events(self):
        """
        Assert that number of concurent events have are available for all labels and equal to particular values
        """

        num_conc_events = num_concurrent_events(self.price_bars.index, self.samples_info_sets['t1'],
                                                self.samples_info_sets.index)
        # Assert for each label we have concurrency value
        self.assertTrue(num_conc_events[self.samples_info_sets.index].shape[0] == self.samples_info_sets.shape[0])
        self.assertTrue(num_conc_events.value_counts()[0] == 5)  # Hours between 14 and 20 label
        self.assertTrue(num_conc_events.value_counts()[1] == 11)
        self.assertTrue(num_conc_events.value_counts()[2] == 5)
        self.assertTrue(num_conc_events.value_counts()[3] == 1)  # Label # 10

    def test_get_av_uniqueness(self):
        """
        Assert that average event uniqueness is available for all labels and equals to particular values
        """

        av_un = get_av_uniqueness_from_triple_barrier(self.samples_info_sets, self.price_bars, num_threads=4)
        # Assert for each label we have uniqueness value
        self.assertTrue(av_un.shape[0] == self.samples_info_sets.shape[0])
        self.assertAlmostEqual(av_un['tW'].iloc[0], 0.66, delta=1e-2)
        self.assertAlmostEqual(av_un['tW'].iloc[2], 0.83, delta=1e-2)
        self.assertAlmostEqual(av_un['tW'].iloc[5], 0.44, delta=1e-2)
        self.assertAlmostEqual(av_un['tW'].iloc[-1], 1.0, delta=1e-2)

    def test_seq_bootstrap(self):
        """
        Test sequential bootstrapping length, indicator matrix length and NaN checks
        """

        ind_mat = get_ind_matrix(self.samples_info_sets.t1, self.price_bars)

        label_endtime = self.samples_info_sets.t1
        trimmed_price_bars_index = self.price_bars[(self.price_bars.index >= self.samples_info_sets.index.min()) &
                                                   (self.price_bars.index <= self.samples_info_sets.t1.max())].index
        bar_index = list(self.samples_info_sets.index)  # Generate index for indicator matrix from t1 and index
        bar_index.extend(self.samples_info_sets.t1)
        bar_index.extend(trimmed_price_bars_index)
        bar_index = sorted(list(set(bar_index)))  # Drop duplicates and sort
        ind_mat_book_implementation = book_ind_mat_implementation(bar_index, label_endtime)

        self.assertTrue(bool((ind_mat_book_implementation.values == ind_mat).all()) is True)
        # Indicator matrix shape should be (unique(meta_label_index+t1+price_bars_index), t1)
        self.assertTrue(ind_mat.shape == (22, 8))

        # Check indicator matrix values for specific labels
        self.assertTrue(bool((ind_mat[:3, 0] == np.ones(3)).all()) is True)
        self.assertTrue(bool((ind_mat[1:4, 1] == np.ones(3)).all()) is True)
        self.assertTrue(bool((ind_mat[4:7, 2] == np.ones(3)).all()) is True)
        self.assertTrue(bool((ind_mat[14:, 6] == np.zeros(8)).all()) is True)

        bootstrapped_samples = seq_bootstrap(ind_mat, compare=False, verbose=True, warmup_samples=None)
        bootstrapped_samples_1000 = seq_bootstrap(ind_mat, compare=True, sample_length=100)
        self.assertTrue(len(bootstrapped_samples) == self.samples_info_sets.shape[0])
        self.assertTrue(len(bootstrapped_samples_1000) == 100)

        # Test sequential bootstrapping on example from a book
        ind_mat = pd.DataFrame(index=range(0, 6), columns=range(0, 3))
        ind_mat.loc[:, 0] = [1, 1, 1, 0, 0, 0]
        ind_mat.loc[:, 1] = [0, 0, 1, 1, 0, 0]
        ind_mat.loc[:, 2] = [0, 0, 0, 0, 1, 1]
        ind_mat = ind_mat.values

        seq_bootstrap(ind_mat, sample_length=3, verbose=True, warmup_samples=[1])  # Show printed probabilities

        # Perform Monte-Carlo test
        standard_unq_array = np.zeros(1000) * np.nan
        seq_unq_array = np.zeros(1000) * np.nan
        for i in range(0, 1000):
            bootstrapped_samples = seq_bootstrap(ind_mat, sample_length=3)
            random_samples = np.random.choice(ind_mat.shape[1], size=3)

            random_unq = get_ind_mat_average_uniqueness(ind_mat[:, random_samples])
            sequential_unq = get_ind_mat_average_uniqueness(ind_mat[:, bootstrapped_samples])

            standard_unq_array[i] = random_unq
            seq_unq_array[i] = sequential_unq

        self.assertTrue(np.mean(seq_unq_array) >= np.mean(standard_unq_array))
        self.assertTrue(np.median(seq_unq_array) >= np.median(standard_unq_array))

    def test_get_ind_mat_av_uniqueness(self):
        """
        Tests get_ind_mat_average_uniqueness function using indicator matrix from the book example
        """

        ind_mat = pd.DataFrame(index=range(0, 6), columns=range(0, 3))
        ind_mat.loc[:, 0] = [1, 1, 1, 0, 0, 0]
        ind_mat.loc[:, 1] = [0, 0, 1, 1, 0, 0]
        ind_mat.loc[:, 2] = [0, 0, 0, 0, 1, 1]
        ind_mat = ind_mat.values

        labels_av_uniqueness = get_ind_mat_average_uniqueness(ind_mat)
        self.assertTrue(abs(labels_av_uniqueness - 0.8571) <= 1e-4)  # Test matrix av.uniqueness

    def test_get_ind_mat_uniqueness(self):
        """
        Tests get_ind_mat_uniqueness function using indicator matrix from the book example
        """

        ind_mat = pd.DataFrame(index=range(0, 6), columns=range(0, 3))
        ind_mat.loc[:, 0] = [1, 1, 1, 0, 0, 0]
        ind_mat.loc[:, 1] = [0, 0, 1, 1, 0, 0]
        ind_mat.loc[:, 2] = [0, 0, 0, 0, 1, 1]
        ind_mat = ind_mat.values

        labels_av_uniqueness = get_ind_mat_label_uniqueness(ind_mat)
        first_sample_unq = labels_av_uniqueness[0]
        second_sample_unq = labels_av_uniqueness[1]
        third_sample_unq = labels_av_uniqueness[2]

        self.assertTrue(abs(first_sample_unq[first_sample_unq > 0].mean() - 0.8333) <= 1e-4)
        self.assertTrue(abs(second_sample_unq[second_sample_unq > 0].mean() - 0.75) <= 1e-4)
        self.assertTrue(abs(third_sample_unq[third_sample_unq > 0].mean() - 1.0) <= 1e-4)
        # Test matrix av.uniqueness
        self.assertTrue(abs(labels_av_uniqueness[labels_av_uniqueness > 0].mean() - 0.8571) <= 1e-4)

    def test_bootstrap_loop_run(self):
        """
        Test one loop iteration of Sequential Bootstrapping
        """
        ind_mat = pd.DataFrame(index=range(0, 6), columns=range(0, 3))
        ind_mat.loc[:, 0] = [1, 1, 1, 0, 0, 0]
        ind_mat.loc[:, 1] = [0, 0, 1, 1, 0, 0]
        ind_mat.loc[:, 2] = [0, 0, 0, 0, 1, 1]
        ind_mat = ind_mat.values

        prev_concurrency = np.zeros(ind_mat.shape[0])

        first_iteration = _bootstrap_loop_run(ind_mat, prev_concurrency)
        self.assertTrue((first_iteration == np.array([1.0, 1.0, 1.0])).all())  # First iteration should always yield 1

        prev_concurrency += ind_mat[:, 1]  # Repeat example from the book
        second_iteration = _bootstrap_loop_run(ind_mat, prev_concurrency)
        second_iteration_prob = second_iteration / second_iteration.sum()

        self.assertTrue(abs((second_iteration_prob - np.array([0.35714286, 0.21428571, 0.42857143])).sum()) <= 1e-8)

    def test_value_error_raise(self):
        """
        Test seq_bootstrap and ind_matrix functions for raising ValueError on nan values
        """
        nan_samples_info_sets = self.samples_info_sets.copy()
        nan_samples_info_sets.loc[pd.Timestamp(2019, 1, 1), 't1'] = None
        with self.assertRaises(ValueError):
            get_ind_matrix(nan_samples_info_sets.t1, self.price_bars)



"""
Third generation models implementation (VPIN)
"""
import pandas as pd


def get_vpin(volume: pd.Series, buy_volume: pd.Series, window: int = 1) -> pd.Series:
    """
    Advances in Financial Machine Learning, p. 292-293.

    Get Volume-Synchronized Probability of Informed Trading (VPIN) from bars

    :param volume: (pd.Series) Bar volume
    :param buy_volume: (pd.Series) Bar volume classified as buy (either tick rule, BVC or aggressor side methods applied)
    :param window: (int) Estimation window
    :return: (pd.Series) VPIN series
    """
    sell_volume = volume - buy_volume
    volume_imbalance = abs(buy_volume - sell_volume)
    return volume_imbalance.rolling(window=window).mean() / volume



"""
Advances in Financial Machine Learning, Marcos Lopez de Prado
Chapter 2: Financial Data Structures

Time bars generation logic
"""

# Imports
from typing import Union, Iterable, Optional
import numpy as np
import pandas as pd

from famlafl.data_structures.base_bars import BaseBars


# pylint: disable=too-many-instance-attributes
class TimeBars(BaseBars):
    """
    Contains all of the logic to construct the time bars. This class shouldn't be used directly.
    Use get_time_bars instead
    """

    def __init__(self, resolution: str, num_units: int, batch_size: int = 20000000):
        """
        Constructor

        :param resolution: (str) Type of bar resolution: ['D', 'H', 'MIN', 'S']
        :param num_units: (int) Number of days, minutes, etc.
        :param batch_size: (int) Number of rows to read in from the csv, per batch
        """
        BaseBars.__init__(self, metric=None, batch_size=batch_size)

        # Threshold at which to sample (in seconds)
        self.time_bar_thresh_mapping = {'D': 86400, 'H': 3600, 'MIN': 60, 'S': 1}  # Number of seconds
        assert resolution in self.time_bar_thresh_mapping, "{} resolution is not implemented".format(resolution)
        self.resolution = resolution  # Type of bar resolution: 'D', 'H', 'MIN', 'S'
        self.num_units = num_units  # Number of days/minutes/...
        self.threshold = self.num_units * self.time_bar_thresh_mapping[self.resolution]
        self.timestamp = None  # Current bar timestamp

    def _reset_cache(self):
        """
        Implementation of abstract method _reset_cache for time bars
        """
        self.open_price = None
        self.close_price = None
        self.high_price, self.low_price = -np.inf, np.inf
        self.cum_statistics = {'cum_ticks': 0, 'cum_dollar_value': 0, 'cum_volume': 0, 'cum_buy_volume': 0}

    def _extract_bars(self, data: Union[list, tuple, np.ndarray]) -> list:
        """
        For loop which compiles time bars.
        We did investigate the use of trying to solve this in a vectorised manner but found that a For loop worked well.

        :param data: (tuple) Contains 3 columns - date_time, price, and volume.
        :return: (list) Extracted bars
        """

        # Iterate over rows
        list_bars = []

        for row in data:
            # Set variables
            date_time = row[0].timestamp()  # Convert to UTC timestamp
            self.tick_num += 1
            price = float(row[1])
            volume = row[2]
            dollar_value = price * volume
            signed_tick = self._apply_tick_rule(price)

            timestamp_threshold = (int(
                float(date_time)) // self.threshold + 1) * self.threshold  # Current tick boundary timestamp

            # Init current bar timestamp with first ticks boundary timestamp
            if self.timestamp is None:
                self.timestamp = timestamp_threshold
            # Bar generation condition
            # Current ticks bar timestamp differs from current bars timestamp
            elif self.timestamp < timestamp_threshold:
                self._create_bars(self.timestamp, self.close_price,
                                  self.high_price, self.low_price, list_bars)

                # Reset cache
                self._reset_cache()
                self.timestamp = timestamp_threshold  # Current bar timestamp update

            # Update counters
            if self.open_price is None:
                self.open_price = price

            # Update high low prices
            self.high_price, self.low_price = self._update_high_low(price)

            # Update close price
            self.close_price = price

            # Calculations
            self.cum_statistics['cum_ticks'] += 1
            self.cum_statistics['cum_dollar_value'] += dollar_value
            self.cum_statistics['cum_volume'] += volume
            if signed_tick == 1:
                self.cum_statistics['cum_buy_volume'] += volume

        return list_bars


def get_time_bars(file_path_or_df: Union[str, Iterable[str], pd.DataFrame], resolution: str = 'D', num_units: int = 1, batch_size: int = 20000000,
                  verbose: bool = True, to_csv: bool = False, output_path: Optional[str] = None):
    """
    Creates Time Bars: date_time, open, high, low, close, volume, cum_buy_volume, cum_ticks, cum_dollar_value.

    :param file_path_or_df: (str, iterable of str, or pd.DataFrame) Path to the csv file(s) or Pandas Data Frame containing raw tick data
                            in the format[date_time, price, volume]
    :param resolution: (str) Resolution type ('D', 'H', 'MIN', 'S')
    :param num_units: (int) Number of resolution units (3 days for example, 2 hours)
    :param batch_size: (int) The number of rows per batch. Less RAM = smaller batch size.
    :param verbose: (int) Print out batch numbers (True or False)
    :param to_csv: (bool) Save bars to csv after every batch run (True or False)
    :param output_path: (str) Path to csv file, if to_csv is True
    :return: (pd.DataFrame) Dataframe of time bars, if to_csv=True return None
    """

    bars = TimeBars(resolution=resolution, num_units=num_units, batch_size=batch_size)
    time_bars = bars.batch_run(file_path_or_df=file_path_or_df, verbose=verbose, to_csv=to_csv, output_path=output_path)
    return time_bars



"""
Module which implements feature PCA compression and PCA analysis of feature importance.
"""

import pandas as pd
import numpy as np
from scipy.stats import weightedtau, kendalltau, spearmanr, pearsonr


def _get_eigen_vector(dot_matrix, variance_thresh, num_features=None):
    """
    Advances in Financial Machine Learning, Snippet 8.5, page 119.

    Computation of Orthogonal Features

    Gets eigen values and eigen vector from matrix which explain % variance_thresh of total variance.

    :param dot_matrix: (np.array): Matrix for which eigen values/vectors should be computed.
    :param variance_thresh: (float): Percentage % of overall variance which compressed vectors should explain.
    :param num_features: (int) Manually set number of features, overrides variance_thresh. (None by default)
    :return: (pd.Series, pd.DataFrame): Eigenvalues, Eigenvectors.
    """
    # Compute eigen_vec from dot prod matrix, reduce dimension
    eigen_val, eigen_vec = np.linalg.eigh(dot_matrix)
    idx = eigen_val.argsort()[::-1]  # Arguments for sorting eigen_val desc
    eigen_val, eigen_vec = eigen_val[idx], eigen_vec[:, idx]

    # 2) Only positive eigen_vals
    eigen_val = pd.Series(eigen_val, index=['PC_' + str(i + 1) for i in range(eigen_val.shape[0])])
    eigen_vec = pd.DataFrame(eigen_vec, index=dot_matrix.index, columns=eigen_val.index)
    eigen_vec = eigen_vec.loc[:, eigen_val.index]

    # 3) Reduce dimension, form PCs
    cum_var = eigen_val.cumsum() / eigen_val.sum()
    if num_features is not None: # Either a preset number of features
        dim = num_features - 1
    else: # Or defined by variance level
        dim = cum_var.values.searchsorted(variance_thresh)
    eigen_val, eigen_vec = eigen_val.iloc[:dim + 1], eigen_vec.iloc[:, :dim + 1]
    return eigen_val, eigen_vec


def _standardize_df(data_frame):
    """
    Helper function which divides df by std and extracts mean.

    :param data_frame: (pd.DataFrame): Dataframe to standardize
    :return: (pd.DataFrame): Standardized dataframe
    """
    return data_frame.sub(data_frame.mean(), axis=1).div(data_frame.std(), axis=1)


def get_orthogonal_features(feature_df, variance_thresh=.95, num_features=None):
    """
    Advances in Financial Machine Learning, Snippet 8.5, page 119.

    Computation of Orthogonal Features.

    Gets PCA orthogonal features.

    :param feature_df: (pd.DataFrame): Dataframe of features.
    :param variance_thresh: (float): Percentage % of overall variance which compressed vectors should explain.
    :param num_features: (int) Manually set number of features, overrides variance_thresh. (None by default)
    :return: (pd.DataFrame): Compressed PCA features which explain %variance_thresh of variance.
    """
    # Given a dataframe of features, compute orthogonal features
    feature_df_standard = _standardize_df(feature_df)  # Standardize
    dot_matrix = pd.DataFrame(np.dot(feature_df_standard.T, feature_df_standard), index=feature_df.columns,
                              columns=feature_df.columns)
    _, eigen_vec = _get_eigen_vector(dot_matrix, variance_thresh, num_features)
    pca_features = np.dot(feature_df_standard, eigen_vec)
    return pca_features


def get_pca_rank_weighted_kendall_tau(feature_imp, pca_rank):
    """
    Advances in Financial Machine Learning, Snippet 8.6, page 121.

    Computes Weighted Kendall's Tau Between Feature Importance and Inverse PCA Ranking.

    :param feature_imp: (np.array): Feature mean importance.
    :param pca_rank: (np.array): PCA based feature importance rank.
    :return: (float): Weighted Kendall Tau of feature importance and inverse PCA rank with p_value.
    """
    return weightedtau(feature_imp, pca_rank ** -1.0)


def feature_pca_analysis(feature_df, feature_importance, variance_thresh=0.95):
    """
    Performs correlation analysis between feature importance (MDI for example, supervised) and PCA eigenvalues
    (unsupervised).

    High correlation means that probably the pattern identified by the ML algorithm is not entirely overfit.

    :param feature_df: (pd.DataFrame): Features dataframe.
    :param feature_importance: (pd.DataFrame): Individual MDI feature importance.
    :param variance_thresh: (float): Percentage % of overall variance which compressed vectors should explain in PCA compression.
    :return: (dict): Dictionary with kendall, spearman, pearson and weighted_kendall correlations and p_values.
    """
    feature_df_standard = _standardize_df(feature_df)  # Standardize
    dot = pd.DataFrame(np.dot(feature_df_standard.T, feature_df_standard), index=feature_df.columns,
                       columns=feature_df.columns)
    eigen_val, eigen_vec = _get_eigen_vector(dot, variance_thresh)

    # Compute correlations between eigen values for each eigen vector vs mdi importance
    all_eigen_values = []  # All eigen values in eigen vectors
    corr_dict = {'Pearson': [], 'Spearman': [], 'Kendall': []}  # Dictionary containing correlation metrics
    for vec in eigen_vec.columns:
        all_eigen_values.extend(abs(eigen_vec[vec].values * eigen_val[vec]))

    # We need to repeat importance array # of eigen vector times to generate correlation for all_eigen_values
    repeated_importance_array = np.tile(feature_importance['mean'].values, len(eigen_vec.columns))

    for corr_type, function in zip(corr_dict.keys(), [pearsonr, spearmanr, kendalltau]):
        corr_coef = function(repeated_importance_array, all_eigen_values)
        corr_dict[corr_type] = corr_coef

    # Get Rank based weighted Tau correlation
    feature_pca_rank = (eigen_val * eigen_vec).abs().sum(axis=1).rank(
        ascending=False)  # Sum of absolute values across all eigen vectors
    corr_dict['Weighted_Kendall_Rank'] = get_pca_rank_weighted_kendall_tau(feature_importance['mean'].values,
                                                                           feature_pca_rank.values)
    return corr_dict



"""
Initialize Online Portfolio Selection Module.
"""
# Parent Class
from famlafl.online_portfolio_selection.base import OLPS
from famlafl.online_portfolio_selection.up import UP

# Benchmarks
from famlafl.online_portfolio_selection.bah import BAH
from famlafl.online_portfolio_selection.best_stock import BestStock
from famlafl.online_portfolio_selection.crp import CRP
from famlafl.online_portfolio_selection.bcrp import BCRP

# Momentum
from famlafl.online_portfolio_selection.eg import EG
from famlafl.online_portfolio_selection.ftl import FTL
from famlafl.online_portfolio_selection.ftrl import FTRL

# Mean Reversion
from famlafl.online_portfolio_selection.rmr import RMR
from famlafl.online_portfolio_selection.pamr import PAMR
from famlafl.online_portfolio_selection.cwmr import CWMR
from famlafl.online_portfolio_selection.olmar import OLMAR

# Pattern Matching
from famlafl.online_portfolio_selection.corn import CORN
from famlafl.online_portfolio_selection.cornu import CORNU
from famlafl.online_portfolio_selection.cornk import CORNK
from famlafl.online_portfolio_selection.scorn import SCORN
from famlafl.online_portfolio_selection.scornk import SCORNK
from famlafl.online_portfolio_selection.fcorn import FCORN
from famlafl.online_portfolio_selection.fcornk import FCORNK



"""
Explosiveness tests: SADF
"""

from typing import Union, Tuple
import pandas as pd
import numpy as np
from famlafl.util.multiprocess import mp_pandas_obj


# pylint: disable=invalid-name

def _get_sadf_at_t(X: pd.DataFrame, y: pd.DataFrame, min_length: int, model: str, phi: float) -> float:
    """
    Advances in Financial Machine Learning, Snippet 17.2, page 258.

    SADF's Inner Loop (get SADF value at t)

    :param X: (pd.DataFrame) Lagged values, constants, trend coefficients
    :param y: (pd.DataFrame) Y values (either y or y.diff())
    :param min_length: (int) Minimum number of samples needed for estimation
    :param model: (str) Either 'linear', 'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :param phi: (float) Coefficient to penalize large sample lengths when computing SMT, in [0, 1]
    :return: (float) SADF statistics for y.index[-1]
    """
    start_points, bsadf = range(0, y.shape[0] - min_length + 1), -np.inf
    for start in start_points:
        y_, X_ = y[start:], X[start:]
        b_mean_, b_std_ = get_betas(X_, y_)
        if not np.isnan(b_mean_[0]):
            b_mean_, b_std_ = b_mean_[0, 0], b_std_[0, 0] ** 0.5
            # TODO: Rewrite logic of this module to avoid division by zero
            with np.errstate(invalid='ignore'):
                all_adf = b_mean_ / b_std_
            if model[:2] == 'sm':
                all_adf = np.abs(all_adf) / (y.shape[0]**phi)
            if all_adf > bsadf:
                bsadf = all_adf
    return bsadf


def _get_y_x(series: pd.Series, model: str, lags: Union[int, list],
             add_const: bool) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """
    Advances in Financial Machine Learning, Snippet 17.2, page 258-259.

    Preparing The Datasets

    :param series: (pd.Series) Series to prepare for test statistics generation (for example log prices)
    :param model: (str) Either 'linear', 'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :param lags: (int or list) Either number of lags to use or array of specified lags
    :param add_const: (bool) Flag to add constant
    :return: (pd.DataFrame, pd.DataFrame) Prepared y and X for SADF generation
    """
    series = pd.DataFrame(series)
    series_diff = series.diff().dropna()
    x = _lag_df(series_diff, lags).dropna()
    x['y_lagged'] = series.shift(1).loc[x.index]  # add y_(t-1) column
    y = series_diff.loc[x.index]

    if add_const is True:
        x['const'] = 1

    if model == 'linear':
        x['trend'] = np.arange(x.shape[0])  # Add t to the model (0, 1, 2, 3, 4, 5, .... t)
        beta_column = 'y_lagged'  # Column which is used to estimate test beta statistics
    elif model == 'quadratic':
        x['trend'] = np.arange(x.shape[0]) # Add t to the model (0, 1, 2, 3, 4, 5, .... t)
        x['quad_trend'] = np.arange(x.shape[0]) ** 2 # Add t^2 to the model (0, 1, 4, 9, ....)
        beta_column = 'y_lagged'  # Column which is used to estimate test beta statistics
    elif model == 'sm_poly_1':
        y = series.loc[y.index]
        x = pd.DataFrame(index=y.index)
        x['const'] = 1
        x['trend'] = np.arange(x.shape[0])
        x['quad_trend'] = np.arange(x.shape[0]) ** 2
        beta_column = 'quad_trend'
    elif model == 'sm_poly_2':
        y = np.log(series.loc[y.index])
        x = pd.DataFrame(index=y.index)
        x['const'] = 1
        x['trend'] = np.arange(x.shape[0])
        x['quad_trend'] = np.arange(x.shape[0]) ** 2
        beta_column = 'quad_trend'
    elif model == 'sm_exp':
        y = np.log(series.loc[y.index])
        x = pd.DataFrame(index=y.index)
        x['const'] = 1
        x['trend'] = np.arange(x.shape[0])
        beta_column = 'trend'
    elif model == 'sm_power':
        y = np.log(series.loc[y.index])
        x = pd.DataFrame(index=y.index)
        x['const'] = 1
        # TODO: Rewrite logic of this module to avoid division by zero
        with np.errstate(divide='ignore'):
            x['log_trend'] = np.log(np.arange(x.shape[0]))
        beta_column = 'log_trend'
    else:
        raise ValueError('Unknown model')

    # Move y_lagged column to the front for further extraction
    columns = list(x.columns)
    columns.insert(0, columns.pop(columns.index(beta_column)))
    x = x[columns]
    return x, y


def _lag_df(df: pd.DataFrame, lags: Union[int, list]) -> pd.DataFrame:
    """
    Advances in Financial Machine Learning, Snipet 17.3, page 259.

    Apply Lags to DataFrame

    :param df: (int or list) Either number of lags to use or array of specified lags
    :param lags: (int or list) Lag(s) to use
    :return: (pd.DataFrame) Dataframe with lags
    """
    df_lagged = pd.DataFrame()
    if isinstance(lags, int):
        lags = range(1, lags + 1)
    else:
        lags = [int(lag) for lag in lags]

    for lag in lags:
        temp_df = df.shift(lag).copy(deep=True)
        temp_df.columns = [str(i) + '_' + str(lag) for i in temp_df.columns]
        df_lagged = df_lagged.join(temp_df, how='outer')
    return df_lagged


def get_betas(X: pd.DataFrame, y: pd.DataFrame) -> Tuple[np.array, np.array]:
    """
    Advances in Financial Machine Learning, Snippet 17.4, page 259.

    Fitting The ADF Specification (get beta estimate and estimate variance)

    :param X: (pd.DataFrame) Features(factors)
    :param y: (pd.DataFrame) Outcomes
    :return: (np.array, np.array) Betas and variances of estimates
    """
    xy = np.dot(X.T, y)
    xx = np.dot(X.T, X)

    try:
        xx_inv = np.linalg.inv(xx)
    except np.linalg.LinAlgError:
        return [np.nan], [[np.nan, np.nan]]

    b_mean = np.dot(xx_inv, xy)
    err = y - np.dot(X, b_mean)
    b_var = np.dot(err.T, err) / (X.shape[0] - X.shape[1]) * xx_inv

    return b_mean, b_var


def _sadf_outer_loop(X: pd.DataFrame, y: pd.DataFrame, min_length: int, model: str, phi: float,
                     molecule: list) -> pd.Series:
    """
    This function gets SADF for t times from molecule

    :param X: (pd.DataFrame) Features(factors)
    :param y: (pd.DataFrame) Outcomes
    :param min_length: (int) Minimum number of observations
    :param model: (str) Either 'linear', 'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :param phi: (float) Coefficient to penalize large sample lengths when computing SMT, in [0, 1]
    :param molecule: (list) Indices to get SADF
    :return: (pd.Series) SADF statistics
    """
    sadf_series = pd.Series(index=molecule, dtype='float64')
    for index in molecule:
        X_subset = X.loc[:index].values
        y_subset = y.loc[:index].values.reshape(-1, 1)
        value = _get_sadf_at_t(X_subset, y_subset, min_length, model, phi)
        sadf_series[index] = value
    return sadf_series


def get_sadf(series: pd.Series, model: str, lags: Union[int, list], min_length: int, add_const: bool = False,
             phi: float = 0, num_threads: int = 8, verbose: bool = True) -> pd.Series:
    """
    Advances in Financial Machine Learning, p. 258-259.

    Multithread implementation of SADF

    SADF fits the ADF regression at each end point t with backwards expanding start points. For the estimation
    of SADF(t), the right side of the window is fixed at t. SADF recursively expands the beginning of the sample
    up to t - min_length, and returns the sup of this set.

    When doing with sub- or super-martingale test, the variance of beta of a weak long-run bubble may be smaller than
    one of a strong short-run bubble, hence biasing the method towards long-run bubbles. To correct for this bias,
    ADF statistic in samples with large lengths can be penalized with the coefficient phi in [0, 1] such that:

    ADF_penalized = ADF / (sample_length ^ phi)

    :param series: (pd.Series) Series for which SADF statistics are generated
    :param model: (str) Either 'linear', 'quadratic', 'sm_poly_1', 'sm_poly_2', 'sm_exp', 'sm_power'
    :param lags: (int or list) Either number of lags to use or array of specified lags
    :param min_length: (int) Minimum number of observations needed for estimation
    :param add_const: (bool) Flag to add constant
    :param phi: (float) Coefficient to penalize large sample lengths when computing SMT, in [0, 1]
    :param num_threads: (int) Number of cores to use
    :param verbose: (bool) Flag to report progress on asynch jobs
    :return: (pd.Series) SADF statistics
    """
    X, y = _get_y_x(series, model, lags, add_const)
    molecule = y.index[min_length:y.shape[0]]

    sadf_series = mp_pandas_obj(func=_sadf_outer_loop,
                                pd_obj=('molecule', molecule),
                                X=X,
                                y=y,
                                min_length=min_length,
                                model=model,
                                phi=phi,
                                num_threads=num_threads,
                                verbose=verbose,
                                )
    return sadf_series



"""
Second generation models features: Kyle lambda, Amihud Lambda, Hasbrouck lambda (bar and trade based)
"""

from typing import List
import numpy as np
import pandas as pd

from famlafl.structural_breaks.sadf import get_betas

# pylint: disable=invalid-name
def get_bar_based_kyle_lambda(close: pd.Series, volume: pd.Series, window: int = 20) -> pd.Series:
    """
    Advances in Financial Machine Learning, p. 286-288.

    Get Kyle lambda from bars data

    :param close: (pd.Series) Close prices
    :param volume: (pd.Series) Bar volume
    :param window: (int) Rolling window used for estimation
    :return: (pd.Series) Kyle lambdas
    """
    close_diff = close.diff()
    close_diff_sign = close_diff.apply(np.sign)
    close_diff_sign = close_diff_sign.replace(0, np.nan).ffill()  # Replace 0 values with previous using forward fill
    volume_mult_trade_signs = volume * close_diff_sign  # bt * Vt
    return (close_diff / volume_mult_trade_signs).rolling(window=window).mean()


def get_bar_based_amihud_lambda(close: pd.Series, dollar_volume: pd.Series, window: int = 20) -> pd.Series:
    """
    Advances in Financial Machine Learning, p.288-289.

    Get Amihud lambda from bars data

    :param close: (pd.Series) Close prices
    :param dollar_volume: (pd.Series) Dollar volumes
    :param window: (int) rolling window used for estimation
    :return: (pd.Series) of Amihud lambda
    """
    returns_abs = np.log(close / close.shift(1)).abs()
    return (returns_abs / dollar_volume).rolling(window=window).mean()


def get_bar_based_hasbrouck_lambda(close: pd.Series, dollar_volume: pd.Series, window: int = 20) -> pd.Series:
    """
    Advances in Financial Machine Learning, p.289-290.

    Get Hasbrouck lambda from bars data

    :param close: (pd.Series) Close prices
    :param dollar_volume: (pd.Series) Dollar volumes
    :param window: (int) Rolling window used for estimation
    :return: (pd.Series) Hasbrouck lambda
    """
    log_ret = np.log(close / close.shift(1))
    log_ret_sign = log_ret.apply(np.sign)
    log_ret_sign = log_ret_sign.replace(0, np.nan).ffill() # Forward fill after replacing zeros

    signed_dollar_volume_sqrt = log_ret_sign * np.sqrt(dollar_volume)
    return (log_ret / signed_dollar_volume_sqrt).rolling(window=window).mean()


def get_trades_based_kyle_lambda(price_diff: list, volume: list, aggressor_flags: list) -> List[float]:
    """
    Advances in Financial Machine Learning, p.286-288.

    Get Kyle lambda from trades data

    :param price_diff: (list) Price diffs
    :param volume: (list) Trades sizes
    :param aggressor_flags: (list) Trade directions [-1, 1]  (tick rule or aggressor side can be used to define)
    :return: (list) Kyle lambda for a bar and t-value
    """
    signed_volume = np.array(volume) * np.array(aggressor_flags)
    X = np.array(signed_volume).reshape(-1, 1)
    y = np.array(price_diff)
    coef, std = get_betas(X, y)
    t_value = coef[0] / std[0] if std[0] > 0 else np.array([0])
    return [coef[0], t_value[0]]


def get_trades_based_amihud_lambda(log_ret: list, dollar_volume: list) -> List[float]:
    """
    Advances in Financial Machine Learning, p.288-289.

    Get Amihud lambda from trades data

    :param log_ret: (list) Log returns
    :param dollar_volume: (list) Dollar volumes (price * size)
    :return: (float) Amihud lambda for a bar
    """
    X = np.array(dollar_volume).reshape(-1, 1)
    y = np.abs(np.array(log_ret))
    coef, std = get_betas(X, y)
    t_value = coef[0] / std[0] if std[0] > 0 else np.array([0])
    return [coef[0], t_value[0]]


def get_trades_based_hasbrouck_lambda(log_ret: list, dollar_volume: list, aggressor_flags: list) -> List[float]:
    """
    Advances in Financial Machine Learning, p.289-290.

    Get Hasbrouck lambda from trades data

    :param log_ret: (list) Log returns
    :param dollar_volume: (list) Dollar volumes (price * size)
    :param aggressor_flags: (list) Trade directions [-1, 1]  (tick rule or aggressor side can be used to define)
    :return: (list) Hasbrouck lambda for a bar and t value
    """
    X = (np.sqrt(np.array(dollar_volume)) * np.array(aggressor_flags)).reshape(-1, 1)
    y = np.abs(np.array(log_ret))
    coef, std = get_betas(X, y)
    t_value = coef[0] / std[0] if std[0] > 0 else np.array([0])
    return [coef[0], t_value[0]]



# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.base import OLPS


class EG(OLPS):
    """
    This class implements the Exponential Gradient Portfolio strategy. It is reproduced with
    modification from the following paper:
    `Li, B., Hoi, S. C.H., 2012. OnLine Portfolio Selection: A Survey. ACM Comput.
    Surv. V, N, Article A (December 2012), 33 pages. <https://arxiv.org/abs/1212.2129>`_

    Exponential gradient strategy tracks the best performing stock in the last period while
    keeping previous portfolio information by using a regularization term.
    """

    def __init__(self, update_rule, eta=0.05):
        """
        Initializes with the designated update rule and eta, the learning rate.

        :param update_rule: (str) 'MU': Multiplicative Update, 'GP': Gradient Projection,
                                  'EM': Expectation Maximization. All three update rules return
                                  similar results with slight differences.
        :param eta: (float) Learning rate with range of [0, inf). Low rate indicates the passiveness
                            of following the momentum and high rate indicates the aggressivness of
                            following the momentum. Depending on the dataset either 0.05 or 20 usually
                            have the highest returns.
        """
        super().__init__()
        self.eta = eta
        self.update_rule = update_rule

    def _update_weight(self, time):
        """
        Predicts the next time's portfolio weight.

        :param time: (int) Current time period.
        :return: (np.array) New portfolio weights using exponential gradient.
        """
        # Get the window's price relative.
        relative_return = self.relative_return[time]

        # Take the dot product of weights and price relative.
        dot_product = np.dot(self.weights, relative_return)

        # Multiplicative update.
        if self.update_rule == 'MU':
            new_weight = self.weights * np.exp(self.eta * relative_return / dot_product)
        elif self.update_rule == 'EM':
            new_weight = self.weights * (1 + self.eta * (relative_return / dot_product - 1))
        elif self.update_rule == 'GP':
            new_weight = self.weights + self.eta * (relative_return - np.ones(self.number_of_assets)
                                                    * np.mean(relative_return) / dot_product)
        else:
            raise ValueError("Incorrect update rule. Try 'MU','EM', or 'GP'.")
        new_weight = self._normalize(new_weight)
        return new_weight



"""
Tests the financial data structures
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.data_structures import standard_data_structures as ds


class TestDataStructures(unittest.TestCase):
    """
    Test the various financial data structures:
    1. Dollar bars
    2. Volume bars
    3. Tick bars
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        self.path = project_path + '/test_data/tick_data.csv'

    def test_dollar_bars(self):
        """
        Tests the dollar bars implementation.
        """
        threshold = 100000

        # Creating a dynamic threshold
        data = pd.read_csv(self.path)
        data.index = pd.to_datetime(data['Date and Time'])
        data = data.drop('Date and Time', axis=1)

        t_constant = pd.Series([100000], index=[data.index[0]])
        t_dynamic = pd.Series([10000, 20000, 50000], index=[data.index[0], data.index[40], data.index[80]])
        t_low = pd.Series([1000], index=[data.index[0]])

        # Generating dollar bars

        db1 = ds.get_dollar_bars(self.path, threshold=threshold, batch_size=1000, verbose=False)
        db2 = ds.get_dollar_bars(self.path, threshold=threshold, batch_size=50, verbose=False)
        db3 = ds.get_dollar_bars(self.path, threshold=threshold, batch_size=10, verbose=False)
        ds.get_dollar_bars(self.path, threshold=threshold, batch_size=50, verbose=False,
                           to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db4.shape == db1.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db4.values == db1.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[0, 'open'] == 1205)
        self.assertTrue(db1.loc[0, 'high'] == 1904.75)
        self.assertTrue(db1.loc[0, 'low'] == 1005.0)
        self.assertTrue(db1.loc[0, 'close'] == 1304.5)

        # Testing dynamic threshold size
        df_constant = ds.get_dollar_bars(self.path, threshold=t_constant, batch_size=1000, verbose=False)
        df_dynamic = ds.get_dollar_bars(self.path, threshold=t_dynamic, batch_size=1000, verbose=False)
        df_low = ds.get_dollar_bars(self.path, threshold=t_low, batch_size=1000, verbose=False)

        # Assert that constant size outputs the same result
        self.assertTrue(df_constant.shape == db1.shape)
        self.assertTrue(np.all(df_constant.values == db1.values))

        # Assert sizes of different thresolds
        self.assertTrue(df_dynamic.shape == (14, 10))
        self.assertTrue(df_low.shape == (99, 10))

        # delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_volume_bars(self):
        """
        Tests the volume bars implementation.
        """
        threshold = 30

        # Creating a dynamic threshold
        data = pd.read_csv(self.path)
        data.index = pd.to_datetime(data['Date and Time'])
        data = data.drop('Date and Time', axis=1)

        t_constant = pd.Series([30], index=[data.index[0]])
        t_dynamic = pd.Series([5, 10, 30], index=[data.index[0], data.index[40], data.index[80]])
        t_low = pd.Series([5], index=[data.index[0]])

        db1 = ds.get_volume_bars(self.path, threshold=threshold, batch_size=1000, verbose=False)
        db2 = ds.get_volume_bars(self.path, threshold=threshold, batch_size=50, verbose=False)
        db3 = ds.get_volume_bars(self.path, threshold=threshold, batch_size=10, verbose=False)
        ds.get_volume_bars(self.path, threshold=threshold, batch_size=50, verbose=False,
                           to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db4.shape == db1.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db4.values == db1.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[0, 'open'] == 1205)
        self.assertTrue(db1.loc[0, 'high'] == 1904.75)
        self.assertTrue(db1.loc[0, 'low'] == 1005.0)
        self.assertTrue(db1.loc[0, 'close'] == 1304.75)

        # Testing dynamic threshold size
        df_constant = ds.get_volume_bars(self.path, threshold=t_constant, batch_size=1000, verbose=False)
        df_dynamic = ds.get_volume_bars(self.path, threshold=t_dynamic, batch_size=1000, verbose=False)
        df_low = ds.get_volume_bars(self.path, threshold=t_low, batch_size=1000, verbose=False)

        # Assert that constant size outputs the same result
        self.assertTrue(df_constant.shape == db1.shape)
        self.assertTrue(np.all(df_constant.values == db1.values))

        # Assert sizes of different thresolds
        self.assertTrue(df_dynamic.shape == (20, 10))
        self.assertTrue(df_low.shape == (32, 10))

        # delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_tick_bars(self):
        """
        Test the tick bars implementation.
        """
        threshold = 10

        # Creating a dynamic threshold
        data = pd.read_csv(self.path)
        data.index = pd.to_datetime(data['Date and Time'])
        data = data.drop('Date and Time', axis=1)

        t_constant = pd.Series([10], index=[data.index[0]])
        t_dynamic = pd.Series([2, 5, 10], index=[data.index[0], data.index[40], data.index[80]])
        t_low = pd.Series([2], index=[data.index[0]])

        db1 = ds.get_tick_bars(self.path, threshold=threshold, batch_size=1000, verbose=False)
        db2 = ds.get_tick_bars(self.path, threshold=threshold, batch_size=50, verbose=False)
        db3 = ds.get_tick_bars(self.path, threshold=threshold, batch_size=10, verbose=False)
        ds.get_tick_bars(self.path, threshold=threshold, batch_size=50, verbose=False,
                         to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db4.shape == db1.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db4.values == db1.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[0, 'open'] == 1205)
        self.assertTrue(db1.loc[0, 'high'] == 1904.75)
        self.assertTrue(db1.loc[0, 'low'] == 1005.0)
        self.assertTrue(db1.loc[0, 'close'] == 1304.50)

        # Testing dynamic threshold size
        df_constant = ds.get_tick_bars(self.path, threshold=t_constant, batch_size=1000, verbose=False)
        df_dynamic = ds.get_tick_bars(self.path, threshold=t_dynamic, batch_size=1000, verbose=False)
        df_low = ds.get_tick_bars(self.path, threshold=t_low, batch_size=1000, verbose=False)

        # Assert that constant size outputs the same result
        self.assertTrue(df_constant.shape == db1.shape)
        self.assertTrue(np.all(df_constant.values == db1.values))

        # Assert sizes of different thresolds
        self.assertTrue(df_dynamic.shape == (28, 10))
        self.assertTrue(df_low.shape == (50, 10))

        # delete generated csv file (if it wasn't generated test would fail)
        os.remove('test.csv')

    def test_multiple_csv_file_input(self):
        """
        Tests that bars generated for multiple csv files and Pandas Data Frame yield the same result
        """
        threshold = 100000

        data = pd.read_csv(self.path)

        idx = int(np.round(len(data) / 2))

        data1 = data.iloc[:idx]
        data2 = data.iloc[idx:]

        tick1 = "tick_data_1.csv"
        tick2 = "tick_data_2.csv"

        data1.to_csv(tick1, index=False)
        data2.to_csv(tick2, index=False)

        file_paths = [tick1, tick2]

        db1 = ds.get_dollar_bars(file_paths, threshold=threshold, batch_size=1000, verbose=False)
        db2 = ds.get_dollar_bars(file_paths, threshold=threshold, batch_size=50, verbose=False)
        db3 = ds.get_dollar_bars(file_paths, threshold=threshold, batch_size=10, verbose=False)
        ds.get_dollar_bars(self.path, threshold=threshold, batch_size=50, verbose=False,
                           to_csv=True, output_path='test.csv')
        db4 = pd.read_csv('test.csv', parse_dates=[0])

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)
        self.assertTrue(db4.shape == db1.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))
        self.assertTrue(np.all(db4.values == db1.values))

        # Assert OHLC is correct
        self.assertTrue(db1.loc[0, 'open'] == 1205)
        self.assertTrue(db1.loc[0, 'high'] == 1904.75)
        self.assertTrue(db1.loc[0, 'low'] == 1005.0)
        self.assertTrue(db1.loc[0, 'close'] == 1304.50)

        # delete generated csv files (if they weren't generated test would fail)
        for csv in (tick1, tick2, "test.csv"):
            os.remove(csv)

    def test_df_as_batch_run_input(self):
        """
        Tests that bars generated for csv file and Pandas Data Frame yield the same result
        """
        threshold = 100000
        tick_data = pd.read_csv(self.path)
        tick_data['Date and Time'] = pd.to_datetime(tick_data['Date and Time'])

        db1 = ds.get_dollar_bars(self.path, threshold=threshold, batch_size=1000, verbose=False)
        ds.get_dollar_bars(self.path, threshold=threshold, batch_size=50, verbose=False,
                           to_csv=True, output_path='test.csv')
        db2 = pd.read_csv('test.csv')
        db2['date_time'] = pd.to_datetime(db2.date_time)
        db3 = ds.get_dollar_bars(tick_data, threshold=threshold, batch_size=10, verbose=False)

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))

    def test_list_as_run_input(self):
        """
        Tests that data generated with csv file and list yield the same result
        """
        threshold = 100000
        tick_data = pd.read_csv(self.path)
        tick_data['Date and Time'] = pd.to_datetime(tick_data['Date and Time'])

        db1 = ds.get_dollar_bars(self.path, threshold=threshold, batch_size=1000, verbose=False)
        ds.get_dollar_bars(self.path, threshold=threshold, batch_size=50, verbose=False,
                           to_csv=True, output_path='test.csv')
        db2 = pd.read_csv('test.csv')
        db2['date_time'] = pd.to_datetime(db2.date_time)

        bars = ds.StandardBars(metric='cum_dollar_value', threshold=threshold)
        cols = ['date_time', 'tick_num', 'open', 'high', 'low', 'close', 'volume', 'cum_buy_volume', 'cum_ticks',
                'cum_dollar_value']

        data = tick_data.values.tolist()
        final_bars = bars.run(data)
        db3 = pd.DataFrame(final_bars, columns=cols)

        # Assert diff batch sizes have same number of bars
        self.assertTrue(db1.shape == db2.shape)
        self.assertTrue(db1.shape == db3.shape)

        # Assert same values
        self.assertTrue(np.all(db1.values == db2.values))
        self.assertTrue(np.all(db1.values == db3.values))

    def test_wrong_batch_input_value_error_raise(self):
        """
        Tests ValueError raise when neither pd.DataFrame nor path to csv file are passed to function call
        """
        with self.assertRaises(ValueError):
            ds.get_dollar_bars(None, threshold=20, batch_size=1000, verbose=False)

    def test_wrong_run_input_value_error_raise(self):
        """
        Tests ValueError raise when neither pd.DataFrame nor path to csv file are passed to function call
        """
        with self.assertRaises(ValueError):
            bars = ds.StandardBars(metric='cum_dollar_value')
            bars.run(None)

    def test_csv_format(self):
        """
        Asserts that the csv data being passed is of the correct format.
        """
        wrong_date = ['2019-41-30', 200.00, np.int64(5)]
        wrong_price = ['2019-01-30', 'asd', np.int64(5)]
        wrong_volume = ['2019-01-30', 200.00, '1.5']
        too_many_cols = ['2019-01-30', 200.00, np.int64(5), 'Limit order', 'B23']

        # pylint: disable=protected-access
        self.assertRaises(ValueError, ds.StandardBars._assert_csv,
                          pd.DataFrame(wrong_date).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.StandardBars._assert_csv,
                          pd.DataFrame(too_many_cols).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.StandardBars._assert_csv,
                          pd.DataFrame(wrong_price).T)
        # pylint: disable=protected-access
        self.assertRaises(AssertionError,
                          ds.StandardBars._assert_csv,
                          pd.DataFrame(wrong_volume).T)



"""
Filters are used to filter events based on some kind of trigger. For example a structural break filter can be
used to filter events where a structural break occurs. This event is then used to measure the return from the event
to some event horizon, say a day.
"""

import numpy as np
import pandas as pd


# Snippet 2.4, page 39, The Symmetric CUSUM Filter.
def cusum_filter(raw_time_series, threshold, time_stamps=True):
    """
    Advances in Financial Machine Learning, Snippet 2.4, page 39.

    The Symmetric Dynamic/Fixed CUSUM Filter.

    The CUSUM filter is a quality-control method, designed to detect a shift in the mean value of a measured quantity
    away from a target value. The filter is set up to identify a sequence of upside or downside divergences from any
    reset level zero. We sample a bar t if and only if S_t >= threshold, at which point S_t is reset to 0.

    One practical aspect that makes CUSUM filters appealing is that multiple events are not triggered by raw_time_series
    hovering around a threshold level, which is a flaw suffered by popular market signals such as Bollinger Bands.
    It will require a full run of length threshold for raw_time_series to trigger an event.

    Once we have obtained this subset of event-driven bars, we will let the ML algorithm determine whether the occurrence
    of such events constitutes actionable intelligence. Below is an implementation of the Symmetric CUSUM filter.

    Note: As per the book this filter is applied to closing prices but we extended it to also work on other
    time series such as volatility.

    :param raw_time_series: (pd.Series) Close prices (or other time series, e.g. volatility).
    :param threshold: (float or pd.Series) When the abs(change) is larger than the threshold, the function captures
                      it as an event, can be dynamic if threshold is pd.Series
    :param time_stamps: (bool) Default is to return a DateTimeIndex, change to false to have it return a list.
    :return: (datetime index vector) Vector of datetimes when the events occurred. This is used later to sample.
    """

    t_events = []
    s_pos = 0
    s_neg = 0

    # log returns
    raw_time_series = pd.DataFrame(raw_time_series)  # Convert to DataFrame
    raw_time_series.columns = ['price']
    raw_time_series['log_ret'] = raw_time_series.price.apply(np.log).diff()
    if isinstance(threshold, (float, int)):
        raw_time_series['threshold'] = threshold
    elif isinstance(threshold, pd.Series):
        raw_time_series.loc[threshold.index, 'threshold'] = threshold
    else:
        raise ValueError('threshold is neither float nor pd.Series!')

    raw_time_series = raw_time_series.iloc[1:]  # Drop first na values

    # Get event time stamps for the entire series
    for tup in raw_time_series.itertuples():
        thresh = tup.threshold
        pos = float(s_pos + tup.log_ret)
        neg = float(s_neg + tup.log_ret)
        s_pos = max(0.0, pos)
        s_neg = min(0.0, neg)

        if s_neg < -thresh:
            s_neg = 0
            t_events.append(tup.Index)

        elif s_pos > thresh:
            s_pos = 0
            t_events.append(tup.Index)

    # Return DatetimeIndex or list
    if time_stamps:
        event_timestamps = pd.DatetimeIndex(t_events)
        return event_timestamps

    return t_events


def z_score_filter(raw_time_series, mean_window, std_window, z_score=3, time_stamps=True):
    """
    Filter which implements z_score filter
    (https://stackoverflow.com/questions/22583391/peak-signal-detection-in-realtime-timeseries-data)

    :param raw_time_series: (pd.Series) Close prices (or other time series, e.g. volatility).
    :param mean_window: (int): Rolling mean window
    :param std_window: (int): Rolling std window
    :param z_score: (float): Number of standard deviations to trigger the event
    :param time_stamps: (bool) Default is to return a DateTimeIndex, change to false to have it return a list.
    :return: (datetime index vector) Vector of datetimes when the events occurred. This is used later to sample.
    """
    t_events = raw_time_series[raw_time_series >= raw_time_series.rolling(window=mean_window).mean() +
                               z_score * raw_time_series.rolling(window=std_window).std()].index
    if time_stamps:
        event_timestamps = pd.DatetimeIndex(t_events)
        return event_timestamps
    return t_events



"""
Implements general backtest statistics
"""
from famlafl.backtest_statistics.backtests import CampbellBacktesting
from famlafl.backtest_statistics.statistics import (timing_of_flattening_and_flips, average_holding_period,
                                                     bets_concentration, all_bets_concentration,
                                                     drawdown_and_time_under_water, sharpe_ratio,
                                                     information_ratio, probabilistic_sharpe_ratio,
                                                     deflated_sharpe_ratio, minimum_track_record_length)



# pylint: disable=protected-access
"""
Tests the Nested Clustered Optimization (NCO) algorithm.
"""

import unittest
import numpy as np
import pandas as pd
from famlafl.portfolio_optimization.nco import NCO


class TestNCO(unittest.TestCase):
    """
    Tests different functions of the NCO algorithm class.
    """

    def setUp(self):
        """
        Initialize
        """

    @staticmethod
    def test_simulate_covariance():
        """
        Test the deriving an empirical vector of means and an empirical covariance matrix.
        """

        nco = NCO()

        # Real values used for deriving empirical ones
        mu_vec = np.array([0, 0.1, 0.2, 0.3])
        cov_mat = np.array([[1, 0.1, 0.2, 0.3],
                            [0.1, 1, 0.1, 0.2],
                            [0.2, 0.1, 1, 0.1],
                            [0.3, 0.2, 0.1, 1]])
        num_obs = 100000

        # Using the function
        mu_empir, cov_empir = nco._simulate_covariance(mu_vec, cov_mat, num_obs, False)

        # Testing that with a high number of observations empirical values are close to real ones
        np.testing.assert_almost_equal(mu_empir.flatten(), mu_vec.flatten(), decimal=2)
        np.testing.assert_almost_equal(cov_mat, cov_empir, decimal=2)

        # Also testing the Ledoit-Wolf shrinkage
        mu_empir_shr, cov_empir_shr = nco._simulate_covariance(mu_vec, cov_mat, num_obs, True)
        np.testing.assert_almost_equal(mu_empir_shr.flatten(), mu_vec.flatten(), decimal=2)
        np.testing.assert_almost_equal(cov_mat, cov_empir_shr, decimal=2)

    def test_cluster_kmeans_base(self):
        """
        Test the finding of the optimal partition of clusters using K-Means algorithm.
        """

        nco = NCO()

        # Correlation matrix and parameters used in the K-Means algorithm.
        np.random.seed(0)
        corr_matrix = pd.DataFrame([[1, 0.1, -0.1],
                                    [0.1, 1, -0.3],
                                    [-0.1, -0.3, 1]])

        max_num_clusters = 2
        n_init = 10

        # Expected correlation matrix of clustered elements, clusters and, Silhouette Coefficient series
        expected_corr = pd.DataFrame([[1, 0.1, -0.1],
                                      [0.1, 1, -0.3],
                                      [-0.1, - 0.3, 1]])
        expected_clust = {0: [0, 1], 1: [2]}
        expected_silh_coef = pd.Series([0.100834, 0.167626, 0], index=[0, 1, 2])

        # Finding the clustered corresponding values
        corr, clusters, silh_coef = nco._cluster_kmeans_base(corr_matrix, max_num_clusters, n_init)

        # When maximum number of clusters is not predefined
        corr_no_max, _, _ = nco._cluster_kmeans_base(corr_matrix, None, n_init)

        # Testing if the values are right
        self.assertTrue(clusters == expected_clust)
        np.testing.assert_almost_equal(np.array(corr), np.array(expected_corr), decimal=4)
        np.testing.assert_almost_equal(np.array(silh_coef), np.array(expected_silh_coef), decimal=4)
        np.testing.assert_almost_equal(np.array(corr), np.array(corr_no_max), decimal=4)

    @staticmethod
    def test_allocate_cvo():
        """
        Test the estimates of the Convex Optimization Solution (CVO).
        """

        nco = NCO()

        # Covariance matrix and the desired mu vector
        cov_matrix = np.array([[0.01, 0.002, -0.001],
                               [0.002, 0.04, -0.006],
                               [-0.001, -0.006, 0.01]])

        mu_vec = np.array([1, 1, 1]).reshape(-1, 1)

        # Expected weights for minimum variance allocation
        w_expected = np.array([[0.37686939],
                               [0.14257228],
                               [0.48055833]])

        # Finding the optimal weights
        w_cvo = nco.allocate_cvo(cov_matrix, mu_vec=None)

        # Also when manually inputting the vector mu
        w_cvo_mu = nco.allocate_cvo(cov_matrix, mu_vec=mu_vec)

        # Testing if the optimal allocation is right and if the custom mu works
        np.testing.assert_almost_equal(w_cvo, w_expected, decimal=4)
        np.testing.assert_almost_equal(w_cvo, w_cvo_mu, decimal=4)

    @staticmethod
    def test_allocate_nco():
        """
        Test the estimates the optimal allocation using the (NCO) algorithm
        """

        nco = NCO()

        # Covariance matrix and the custom mu vector
        np.random.seed(0)
        cov_matrix = np.array([[0.01, 0.002, -0.001],
                               [0.002, 0.04, -0.006],
                               [-0.001, -0.006, 0.01]])

        mu_vec = np.array([1, 1, 1]).reshape(-1, 1)

        # Expected weights for minimum variance allocation
        w_expected = np.array([[0.43875825],
                               [0.09237016],
                               [0.4688716]])
        max_num_clusters = 2

        # Finding the optimal weights
        w_nco = nco.allocate_nco(cov_matrix, max_num_clusters=max_num_clusters)

        # Finding the optimal weights using the custom mu vector
        w_nco_mu = nco.allocate_nco(cov_matrix, mu_vec=mu_vec, max_num_clusters=max_num_clusters)

        # Testing if the optimal allocation is right and if the custom mu works
        np.testing.assert_almost_equal(w_nco, w_expected, decimal=4)
        np.testing.assert_almost_equal(w_nco, w_nco_mu, decimal=4)

    @staticmethod
    def test_allocate_mcos():
        """
        Test the estimates of the optimal allocation using the Monte Carlo optimization selection
        """

        nco = NCO()

        # Covariance matrix, mean vector and other variables for the method
        np.random.seed(0)
        mu_vec = np.array([0, 0.1, 0.2, 0.3])
        cov_mat = np.array([[1, 0.1, 0.2, 0.3],
                            [0.1, 1, 0.1, 0.2],
                            [0.2, 0.1, 1, 0.1],
                            [0.3, 0.2, 0.1, 1]])
        num_obs = 100
        num_sims = 2
        kde_bwidth = 0.25
        min_var_portf = True
        lw_shrinkage = False

        # Alternative set of values
        min_var_portf_alt = False
        kde_bwidth_alt = 0

        # Expected weights for minimum variance allocation
        # Second line in the below test was changed after v.0.11.3 from [0.257547, 0.265450, 0.242453, 0.234551]
        #                                                          to   [0.273299, 0.242542, 0.241464, 0.242696]
        # due to scikit-learn changing np.random outputs
        w_cvo_expected = pd.DataFrame([[0.249287, 0.256002, 0.242593, 0.252118],
                                       [0.273299, 0.242542, 0.241464, 0.242696]])

        # Second line in the below test was changed after v.0.11.3 from [0.257547, 0.265450, 0.242453, 0.234551]
        #                                                          to   [0.273299, 0.242542, 0.241464, 0.242696]
        # due to scikit-learn changing np.random outputs
        w_nco_expected = pd.DataFrame([[0.248396, 0.243172, 0.250751, 0.257680],
                                       [0.273299, 0.242542, 0.241464, 0.242696]])

        # Expected weights for maximum Sharpe ratio allocation
        # Values in the below test was changed after v.0.11.3 from [[-1.081719, 1.810936, 1.218067, 3.978880]
        #                                                           [-2.431651, 0.594868, -0.210175, 5.117628]]
        #                                                     to   [[-0.128849, -0.326671,  0.870183,  2.020053]
        #                                                           [-3.786126, -0.881858,  1.418773,  3.062546]]
        # due to scikit-learn changing np.random outputs
        w_cvo_sr_expected = pd.DataFrame([[-0.128849, -0.326671, 0.870183, 2.020053],
                                          [-3.786126, -0.881858, 1.418773, 3.062546]])

        # Values in the below test was changed after v.0.11.3 from [[-1.060835, 1.910503, 1.315026, 3.908128]
        #                                                           [-0.937168, 1.886158, -0.389275, 4.884809]]
        #                                                     to   [[-0.204089, -0.050088,  0.912494,  1.983382]
        #                                                           [-3.723231, -1.809242,  1.862001,  2.61035]]
        # due to scikit-learn changing np.random outputs
        w_nco_sr_expected = pd.DataFrame([[-0.204089, -0.050088, 0.912494, 1.983382],
                                          [-3.723231, -1.809242, 1.862001, 2.61035]])

        # Finding the optimal weights for minimum variance
        w_cvo, w_nco = nco.allocate_mcos(mu_vec, cov_mat, num_obs, num_sims, kde_bwidth, min_var_portf, lw_shrinkage)

        # Finding the optimal weights for maximum Sharpe ratio
        w_cvo_sr, w_nco_sr = nco.allocate_mcos(mu_vec, cov_mat, num_obs, num_sims, kde_bwidth_alt, min_var_portf_alt, lw_shrinkage)

        # Testing if the optimal allocation simulations are right
        np.testing.assert_almost_equal(np.array(w_cvo), np.array(w_cvo_expected), decimal=6)
        np.testing.assert_almost_equal(np.array(w_nco), np.array(w_nco_expected), decimal=6)

        np.testing.assert_almost_equal(np.array(w_cvo_sr), np.array(w_cvo_sr_expected), decimal=6)
        np.testing.assert_almost_equal(np.array(w_nco_sr), np.array(w_nco_sr_expected), decimal=6)

    @staticmethod
    def test_estim_errors_mcos():
        """
        Test the computation the true optimal allocation w, and compares that result with the estimated ones by MCOS.
        """

        nco = NCO()

        # Weights from CVO and NCO, and data to estimate the true optimal weights
        np.random.seed(0)
        w_cvo = pd.DataFrame([[0.249287, 0.256002, 0.242593, 0.252118],
                              [0.257547, 0.265450, 0.242453, 0.234551]])

        w_nco = pd.DataFrame([[0.248396, 0.243172, 0.250751, 0.257680],
                              [0.257547, 0.265450, 0.242453, 0.234551]])

        mu_vec = np.array([0, 0.1, 0.2, 0.3])
        cov_mat = np.array([[1, 0.1, 0.2, 0.3],
                            [0.1, 1, 0.1, 0.2],
                            [0.2, 0.1, 1, 0.1],
                            [0.3, 0.2, 0.1, 1]])
        min_var_portf = True

        # Expected errors
        err_cvo_expected = 0.0062604
        err_nco_expected = 0.0111115

        # Finding the errors in estimations
        err_cvo, err_nco = nco.estim_errors_mcos(w_cvo, w_nco, mu_vec, cov_mat, min_var_portf)

        # Testing if the error computation is right
        np.testing.assert_almost_equal(err_cvo, err_cvo_expected, decimal=4)
        np.testing.assert_almost_equal(err_nco, err_nco_expected, decimal=4)

    @staticmethod
    def test_form_block_matrix():
        """
        Test the creation of a block correlation matrix with given parameters.
        """

        nco = NCO()

        # Parameters to build the block correlation matrix with
        num_blocks = 2
        block_size = 2
        block_corr = 0.3

        # Expected corr matrix
        corr_expected = np.array([[1, 0.3, 0, 0],
                                  [0.3, 1, 0, 0],
                                  [0, 0, 1, 0.3],
                                  [0, 0, 0.3, 1]])

        # Finding the errors in estimations
        corr_matrix = nco._form_block_matrix(num_blocks, block_size, block_corr)

        # Testing if the error computation is right
        np.testing.assert_almost_equal(corr_matrix, corr_expected, decimal=4)

    @staticmethod
    def test_form_true_matrix():
        """
        Test the creation of a random vector of means and a random covariance matrix.
        """

        nco = NCO()

        # Parameters to build the vector of means and a covariance matrix
        np.random.seed(0)
        num_blocks = 2
        block_size = 2
        block_corr = 0.3
        std = 0.3

        # Alternative std parameter
        std_alt = None

        # Expected vector of means and covariance matrix
        mu_expected = np.array([[0.5936214],
                                [0.97226796],
                                [0.8602674],
                                [0.00681664]])

        cov_expected = pd.DataFrame([[0.09, 0.027, 0, 0],
                                     [0.027, 0.09, 0, 0],
                                     [0, 0, 0.09, 0.027],
                                     [0, 0, 0.027, 0.09]])

        mu_alt_expected = np.array([[0.34260886],
                                    [0.12059827],
                                    [0.24366456],
                                    [0.17248975]])

        # Finding the random vector of means and covariance matrix
        mu_vec, cov_matrix = nco.form_true_matrix(num_blocks, block_size, block_corr, std)

        # Also when the std is default
        mu_vec_alt, _ = nco.form_true_matrix(num_blocks, block_size, block_corr, std_alt)

        # Testing if the results are right
        np.testing.assert_almost_equal(mu_vec, mu_expected, decimal=4)
        np.testing.assert_almost_equal(np.array(cov_matrix), np.array(cov_expected), decimal=4)

        np.testing.assert_almost_equal(mu_vec_alt, mu_alt_expected, decimal=4)



"""
Tests the backtests of Campbell research - Haircut Sharpe ratio and Profit Hurdle algorithms.
"""
import unittest
import numpy as np
from famlafl.backtest_statistics.backtests import CampbellBacktesting


class TestCampbellBacktesting(unittest.TestCase):
    # pylint: disable=protected-access
    """
    Tests functions of the CampbellBacktesting class.
    """

    def test_haircut_sharpe_ratios_simple_input(self):
        """
        Test the calculation of haircuts with simple inputs
        """

        sample_frequency = 'M'
        num_observations = 120
        sharpe_ratio = 1
        annualized = True
        autocorr_adjusted = False
        rho_a = 0.1
        num_mult_test = 100
        rho = 0.4
        parameters = (sample_frequency, num_observations, sharpe_ratio, annualized,
                      autocorr_adjusted, rho_a, num_mult_test, rho)

        # Avoiding a random output
        np.random.seed(0)

        backtesting = CampbellBacktesting(400)
        haircuts = backtesting.haircut_sharpe_ratios(*parameters)

        # Testing the adjusted p-values as other outputs are calculated from those
        self.assertAlmostEqual(haircuts[0][0], 0.465, delta=1e-2)
        self.assertAlmostEqual(haircuts[0][1], 0.409, delta=1e-2)
        self.assertAlmostEqual(haircuts[0][2], 0.174, delta=1e-2)
        self.assertAlmostEqual(haircuts[0][3], 0.348, delta=1e-2)

    def test_profit_hurdle(self):
        """
        Test the calculation of haircuts with simple inputs
        """

        num_mult_test = 300
        num_obs = 240
        alpha_sig = 0.05
        vol_anu = 0.1
        rho = 0.4
        parameters = (num_mult_test, num_obs, alpha_sig, vol_anu, rho)

        # Avoiding a random output
        np.random.seed(0)

        backtesting = CampbellBacktesting(400)
        p_values = backtesting.profit_hurdle(*parameters)

        # Testing the adjusted p-values as other outputs are calculated from them
        self.assertAlmostEqual(p_values[0], 0.365, delta=1e-2)
        self.assertAlmostEqual(p_values[1], 0.702, delta=1e-2)
        self.assertAlmostEqual(p_values[2], 0.687, delta=1e-2)
        self.assertAlmostEqual(p_values[3], 0.620, delta=1e-2)
        self.assertAlmostEqual(p_values[4], 0.694, delta=1e-2)

    def test_holm_method_returns(self):
        """
        Test the special inputs to Holm method on required monthly returns.
        Particularly, when there are no exceeding p-values in the simulations
        """

        p_values_simulation = [0.001, 0.0011, 0.0012, 0.0013]
        num_mult_test = 4
        alpha_sig = 0.05
        parameters = (p_values_simulation, num_mult_test, alpha_sig)

        # Avoiding a random output
        np.random.seed(0)

        backtesting = CampbellBacktesting(200)
        tstat = backtesting._holm_method_returns(*parameters)

        # Testing the resulting t-statistic
        self.assertEqual(tstat, 1.96)


    def test_bhy_method_returns(self):
        """
        Test the special inputs to BHY method on required monthly returns.
        """

        p_values_simulation = [0.1, 0.11, 0.12, 0.13]
        p_values_simulation_low = [0.001, 0.0011, 0.0012, 0.0013]
        num_mult_test = 4
        alpha_sig = 0.05

        # Avoiding a random output
        np.random.seed(0)

        backtesting = CampbellBacktesting(200)

        # When too few multiple tests
        tstat = backtesting._bhy_method_returns(p_values_simulation, 1, alpha_sig)
        self.assertEqual(tstat, 1.96)

        # If no exceeding p-values
        tstat = backtesting._bhy_method_returns(p_values_simulation, num_mult_test, alpha_sig)
        self.assertEqual(tstat, 1.96)

        # If exceeding value is first
        tstat = round(backtesting._bhy_method_returns(p_values_simulation_low, num_mult_test, alpha_sig), 3)
        self.assertEqual(tstat, 3.216)

    def test_parameter_calculation(self):
        """
        Test the calculation of parameters of HLZ model by correlation
        """

        backtesting = CampbellBacktesting(200)
        rho = [-0.1, 0.1, 0.3, 0.5, 0.7, 0.9, 1.1]
        expected_result = [0.2, 0.1, 0.3, 0.5, 0.7, 0.9, 0.2]
        parameters = []

        # Avoiding a random output
        np.random.seed(0)

        for rho_el in rho:
            parameters.append(round(backtesting._parameter_calculation(rho_el)[0], 2))
        self.assertEqual(expected_result, parameters)

    def test_annualized_sharpe_ratio(self):
        """
        Test the conversion of Sharpe ratio to annualized and adjusted to autocorrelation of returns
        """

        backtesting = CampbellBacktesting(200)
        sampling_frequency = ['D', 'W', 'M', 'Q', 'A', 'N']
        expected_result_autocorr = [0.905, 0.906, 0.912, 0.928, 1.0, 1.0]
        expected_result_annual = [18.974, 7.211, 3.464, 2.0, 1.0, 1.0]
        parameters_autocorr = []

        # Avoiding a random output
        np.random.seed(0)

        # Tests for not adjusted to autocorrelation
        for freq in sampling_frequency:
            sr_annual_adj = backtesting._annualized_sharpe_ratio(sharpe_ratio=1, sampling_frequency=freq, rho=0.1,
                                                                 annualized=True, autocorr_adjusted=False)
            parameters_autocorr.append(round(sr_annual_adj, 3))
        self.assertEqual(expected_result_autocorr, parameters_autocorr)

        parameters_annual = []
        # Test for already adjusted
        for freq in sampling_frequency:
            sr_annual_adj = backtesting._annualized_sharpe_ratio(sharpe_ratio=1, sampling_frequency=freq, rho=0.1,
                                                                 annualized=False, autocorr_adjusted=True)
            parameters_annual.append(round(sr_annual_adj, 3))
        self.assertEqual(expected_result_annual, parameters_annual)

    def test_monthly_observations(self):
        """
        Test the calculation of monthly observations in a sample
        """

        backtesting = CampbellBacktesting(200)
        sampling_frequency = ['D', 'W', 'M', 'Q', 'A', 'N']
        expected_observations = [0.0, 2.0, 10.0, 30.0, 120., 10.0]
        observations = []

        # Avoiding a random output
        np.random.seed(0)

        # Tests for not adjusted to autocorrelation
        for freq in sampling_frequency:
            monthly_obs = backtesting._monthly_observations(10, freq)
            observations.append(monthly_obs)
        self.assertEqual(expected_observations, observations)



"""
Tests Confidence Weighted Mean Reversion.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.cwmr import CWMR


class TestConfidenceWeightedMeanReversion(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Confidence Weighted Mean Reversion class.
    """

    def setUp(self):
        """
        Sets the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_cwmr_solution(self):
        """
        Test the calculation of CWMR with the original method.
        """
        # Initialize CWMR.
        cwmr = CWMR(confidence=0.5, epsilon=0.5, method='var')
        # Allocates asset prices to CWMR.
        cwmr.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(cwmr.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_cwmr_sd_solution(self):
        """
        Test the calculation of CWMR with the second method.
        """
        # Initialize CWMR.
        cwmr = CWMR(confidence=0.5, epsilon=0.5, method='sd')
        # Allocates asset prices to OLMAR.
        cwmr.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(cwmr.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_cwmr_epsilon_error(self):
        """
        Tests ValueError if epsilon is greater than 1 or less than 0.
        """
        # Initialize CWMR.
        cwmr1 = CWMR(confidence=0.5, epsilon=2)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            cwmr1.allocate(self.data)

        # Initialize CWMR.
        cwmr2 = CWMR(confidence=0.5, epsilon=-1)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            cwmr2.allocate(self.data)

    def test_cwmr_confidence_error(self):
        """
        Tests ValueError if confidence is greater than 1 or less than 0.
        """
        # Initialize CWMR.
        cwmr3 = CWMR(confidence=2, epsilon=0.5)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            cwmr3.allocate(self.data)

        # Initialize CWMR.
        cwmr4 = CWMR(confidence=-1, epsilon=0.5)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            cwmr4.allocate(self.data)

    def test_cwmr_method_error(self):
        """
        Tests ValueError if method is not 'sd' or 'var'.
        """
        # Initialize CWMR.
        cwmr5 = CWMR(confidence=0.5, epsilon=0.5, method='normal')
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            cwmr5.allocate(self.data)

    def test_cwmr_weights_solution(self):
        """
        Test the calculation of CWMR with given weights.
        """
        # Set weights.
        weight = np.zeros(self.data.iloc[0].shape)
        weight[0] = 1
        # Initialize CWMR.
        cwmr6 = CWMR(confidence=0.5, epsilon=0.5, method='var')
        # Allocates asset prices to CWMR.
        cwmr6.allocate(self.data, weights=weight, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(cwmr6.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)



"""
Various miscellaneous microstructural features (VWAP, average tick size)
"""

import numpy as np


def vwap(dollar_volume: list, volume: list) -> float:
    """
    Get Volume Weighted Average Price (VWAP).

    :param dollar_volume: (list) Dollar volumes
    :param volume: (list) Trades sizes
    :return: (float) VWAP value
    """
    return sum(dollar_volume) / sum(volume)


def get_avg_tick_size(tick_size_arr: list) -> float:
    """
    Get average tick size in a bar.

    :param tick_size_arr: (list) Trade sizes
    :return: (float) Average trade size
    """
    return np.mean(tick_size_arr)



# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.up import UP
from famlafl.online_portfolio_selection.corn import CORN


class CORNK(UP):
    """
    This class implements the Correlation Driven Nonparametric Learning - K strategy. It is
    reproduced with modification from the following paper:
    `Li, B., Hoi, S.C., & Gopalkrishnan, V. (2011). CORN: Correlation-driven nonparametric
    learning approach for portfolio selection. ACM TIST, 2,
    21:1-21:29. <https://dl.acm.org/doi/abs/10.1145/1961189.1961193>`_

    Correlation Driven Nonparametric Learning - K formulates a number of experts and tracks the
    experts performance over time. Each period, the strategy decides to allocate capital to
    the top-k experts until the previous time period. This strategy takes an ensemble approach to
    the top performing experts.
    """

    def __init__(self, window, rho, k):
        """
        Initializes Correlation Driven Nonparametric Learning - K with the given number of
        windows, rho values, and k experts.

        :param window: (int) Number of windows to look back for similarity sets. Generates
                             experts with range of [1, 2, ..., w]. The window ranges typically work
                             well with shorter terms of [1, 7].
        :param rho: (int) Number of rho values for threshold. Generates experts with range of
                          [0, 1, (rho-1)/rho]. Higher rho values allow for a greater coverage of
                          possible parameters, but it will slow down the calculations. Rho ranges
                          typically work well with [2, 7].
        :param k: (int) Number of top-k experts. K values have range of [1, window * rho]. Higher
                        number of experts gives a higher exposure to a number of strategies. K value of
                        1 or 2 had the best results as some parameters significantly outperform others.
        """
        self.window = window
        self.rho = rho
        self.k = k
        self.number_of_experts = self.window * self.rho
        super().__init__(number_of_experts=self.number_of_experts, weighted='top-k', k=self.k)

    def _initialize(self, asset_prices, weights, resample_by):
        """
        Initializes the important variables for the object.

        :param asset_prices: (pd.DataFrame) Historical asset prices.
        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :param resample_by: (str) Specifies how to resample the prices. 'D' for Day, 'W' for Week,
                                  'M' for Month. The inputs are based on pandas' resample method.
        """
        # Check that window value is an integer.
        if not isinstance(self.window, int):
            raise ValueError("Window value must be an integer.")

        # Check that rho value is an integer.
        if not isinstance(self.rho, int):
            raise ValueError("Rho value must be an integer.")

        # Check that k value is an integer.
        if not isinstance(self.k, int):
            raise ValueError("K value must be an integer.")

        # Check that window value is at least 1.
        if self.window < 1:
            raise ValueError("Window value must be greater than or equal to 1.")

        # Check that rho value is at least 1.
        if self.rho < 1:
            raise ValueError("Rho value must be greater than or equal to 1.")

        # Check that k value is at least 1.
        if self.k < 1:
            raise ValueError("K value must be greater than or equal to 1.")

        # Check that k value is less than window * rho.
        if self.k > self.number_of_experts:
            raise ValueError("K must be less than or equal to window * rho.")

        super(CORNK, self)._initialize(asset_prices,
                                       weights,
                                       resample_by)

    def _generate_experts(self):
        """
        Generates window * rho experts from window of [1, w] and rho of [0, (rho - 1) / rho].
        """
        # Initialize expert parameters.
        self.expert_params = np.zeros((self.number_of_experts, 2))

        # Pointer to iterate through parameter locations.
        pointer = 0

        # Window from 1 to self.window.
        for n_window in range(1, self.window + 1):
            # Rho from 0 to (rho - 1)/rho.
            for n_rho in range(self.rho):
                # Assign experts with parameters (n_window + 1, n_rho/rho).
                self.expert_params[pointer] = [n_window, n_rho/self.rho]
                # Next pointer.
                pointer += 1
        # Assign parameters.
        for exp in range(self.number_of_experts):
            param = self.expert_params[exp]
            self.experts.append(CORN(int(param[0]), param[1]))



"""
Tests the financial data structures
"""

import unittest
import os
import numpy as np
import pandas as pd

from famlafl.multi_product.etf_trick import get_futures_roll_series


class TestETFTrick(unittest.TestCase):
    """
    Test the various financial data structures:
    1. Dollar bars
    2. Volume bars
    3. Tick bars
    """

    def setUp(self):
        """
        Set the file path for the tick data csv
        """
        project_path = os.path.dirname(__file__)
        path = project_path + '/test_data'

        self.open_df_path = '{}/open_df.csv'.format(path)
        self.close_df_path = '{}/close_df.csv'.format(path)

        self.open_df = pd.read_csv(self.open_df_path, usecols=['date', 'spx'])
        self.close_df = pd.read_csv(self.close_df_path, usecols=['date', 'spx'])

        self.open_df.rename(columns={'spx': 'open'}, inplace=True)
        self.close_df.rename(columns={'spx': 'close'}, inplace=True)

    def test_futures_roll(self):
        """
        Tests get_futures_roll function
        """
        combined_df = self.open_df.merge(self.close_df, on='date')
        combined_df['date'] = pd.to_datetime(combined_df.date)
        combined_df.set_index('date', inplace=True)
        roll_dates = {'futures_1': pd.Timestamp(2017, 3, 20),
                      'futures_2': pd.Timestamp(2018, 1, 17)}  # futures roll dates

        combined_df['current_futures'] = np.where(combined_df.index <= roll_dates['futures_1'], 'futures_1',
                                                  np.where(combined_df.index <= roll_dates['futures_2'], 'futures_2',
                                                           'futures_3'))

        gaps_diff_no_backward = get_futures_roll_series(combined_df, 'open', 'close', 'current_futures',
                                                        'current_futures', method='absolute', roll_backward=False)
        gaps_rel_no_backward = get_futures_roll_series(combined_df, 'open', 'close', 'current_futures',
                                                       'current_futures', method='relative', roll_backward=False)
        gaps_diff_with_backward = get_futures_roll_series(combined_df, 'open', 'close', 'current_futures',
                                                          'current_futures', method='absolute', roll_backward=True)
        gaps_rel_with_backward = get_futures_roll_series(combined_df, 'open', 'close', 'current_futures',
                                                         'current_futures', method='relative', roll_backward=True)
        with self.assertRaises(ValueError):
            get_futures_roll_series(combined_df, 'open', 'close', 'current_futures',
                                    'current_futures', method='unknown', roll_backward=True)

        # Test number of gaps (should be 2 => number of unique gaps should be 3 (0 added)
        self.assertTrue(len(gaps_diff_no_backward.unique()) == len(roll_dates) + 1)
        self.assertTrue(len(gaps_rel_no_backward.unique()) == len(roll_dates) + 1)

        # Assert values of difference method Futures Roll Trick
        self.assertTrue(gaps_diff_no_backward.iloc[0] == 0)
        self.assertTrue(gaps_diff_no_backward.iloc[-1] == -1.75)
        self.assertTrue(gaps_diff_with_backward.iloc[0] == 1.75)
        self.assertTrue(gaps_diff_with_backward.iloc[-1] == 0)

        # Assert values of relative method Futures Roll Trick
        self.assertTrue(gaps_rel_no_backward.iloc[0] == 1)
        self.assertTrue(abs(gaps_rel_no_backward.iloc[-1] - 0.999294) < 1e-6)
        self.assertTrue(abs(gaps_rel_with_backward.iloc[0] - 1 / 0.999294) < 1e-6)
        self.assertTrue(gaps_rel_with_backward.iloc[-1] == 1)



"""
Test various methods of generating feature clusters
"""
import unittest

from famlafl.util.generate_dataset import get_classification_data
from famlafl.clustering.feature_clusters import get_feature_clusters

# pylint: disable=invalid-name
class TestFeatureClusters(unittest.TestCase):
    """
    Test get_feature_clusters function
    """

    def setUp(self):
        """
        Create X, y datasets
        """
        self.X, self.y = get_classification_data(40, 5, 30, 1000, sigma=2)

    def test_get_feature_clusters(self):
        """
        Test get_feature_clusters arguments
        """
        #test for different dependence matrix

        clustered_subsets = get_feature_clusters(self.X, dependence_metric='information_variation',
                                                 distance_metric='angular', linkage_method='single',
                                                 n_clusters=2)
        #hierarchical auto clustering
        clustered_subsets_ha = get_feature_clusters(self.X, dependence_metric='linear',
                                                    distance_metric='angular', linkage_method='single',
                                                    n_clusters=None, critical_threshold=0.2)
        #test for optimal number of clusters and  _check_for_low_silhouette_scores
        #since this is done on test dataset so there will be no features with low silhouette score
        #so we will make a feature with some what lower silhouette score (near to zero) and set
        #the threshold higher (0.2) than that. Also we need a feature to trigger the low degree of freedom
        #condition so, we create a series of zero in the datasets
        self.X['R_5c'] = self.X['R_5'] #this feature is add to introduce low DF in the regressor.
        self.X['R_1c'] = self.X['R_1'] #this will trigger the expection of LinAlgError i.e. presence of singular matrix
        clustered_subsets_distance = get_feature_clusters(self.X, dependence_metric='linear',
                                                          distance_metric=None, linkage_method=None,
                                                          n_clusters=None, critical_threshold=0.2)

        #assertions
        #output clusters must be 2
        self.assertEqual(len(clustered_subsets), 2)
        #The ONC should detect somwhere around 5 clusters
        self.assertAlmostEqual(len(clustered_subsets_ha), 5, delta=1)
        self.assertAlmostEqual(len(clustered_subsets_distance), 5, delta=1)

    def test_value_error_raise(self):
        """
        Test get_feature_clusters for invalid number of clusters arguments
        """
        #Number of clusters larger than number of features
        with self.assertRaises(ValueError):
            get_feature_clusters(self.X, dependence_metric='linear',
                                 distance_metric='angular', linkage_method='single',
                                 n_clusters=int(41))



# pylint: disable=missing-module-docstring
import numpy as np
from famlafl.online_portfolio_selection.base import OLPS


class BestStock(OLPS):
    """
    This class implements the Buy and Hold strategy. It is reproduced with modification from the
    following paper:
    `Li, B., Hoi, S. C.H., 2012. OnLine Portfolio Selection: A Survey. ACM Comput.
    Surv. V, N, Article A (December 2012), 33 pages. <https://arxiv.org/abs/1212.2129>`_

    The Best Stock strategy represents the best performing asset over the period in hindsight.
    """

    def _first_weight(self, weights):
        """
        Sets the initial weight to the best performing stock over the entire time period.

        :param weights: (list/np.array/pd.DataFrame) Initial weights set by the user.
        :return: (np.array) Weights that allocate to the best performing asset.
        """
        # Index of stock that increased the most.
        best_idx = np.argmax(self.relative_return.cumprod(axis=0)[-1])
        # Initialize an array of zeros.
        new_weight = np.zeros(self.number_of_assets)
        # Assign one to the best performing asset.
        new_weight[best_idx] = 1
        return new_weight



"""
Implementation of Chu-Stinchcombe-White test
"""

import pandas as pd
import numpy as np
from famlafl.util import mp_pandas_obj


def _get_values_diff(test_type, series, index, ind):
    """
    Gets the difference between two values given a test type.
    :param test_type: (str) Type of the test ['one_sided', 'two_sided']
    :param series: (pd.Series) Series of values
    :param index: (pd.Index) primary index
    :param ind: (pd.Index) secondary index
    :return: (float) Difference between 2 values
    """
    if test_type == 'one_sided':
        values_diff = series.loc[index] - series.loc[ind]
    elif test_type == 'two_sided':
        values_diff = abs(series.loc[index] - series.loc[ind])
    else:
        raise ValueError('Test type is unknown: can be either one_sided or two_sided')

    return values_diff


def _get_s_n_for_t(series: pd.Series, test_type: str, molecule: list) -> pd.Series:
    """
    Get maximum S_n_t value for each value from molecule for Chu-Stinchcombe-White test

    :param series: (pd.Series) Series to get statistics for
    :param test_type: (str): Two-sided or one-sided test
    :param molecule: (list) Indices to get test statistics for
    :return: (pd.Series) Statistics
    """

    s_n_t_series = pd.DataFrame(index=molecule, columns=['stat', 'critical_value'])
    for index in molecule:

        series_t = series.loc[:index]
        squared_diff = series_t.diff().dropna() ** 2
        integer_index = series_t.index.get_loc(index)
        sigma_sq_t = 1 / (integer_index - 1) * sum(squared_diff)

        max_s_n_value = -np.inf
        max_s_n_critical_value = None  # Corresponds to c_alpha[n,t]

        # Indices difference for the last index would yield 0 -> division by zero warning,
        # no need to index last value iteration
        for ind in series_t.index[:-1]:
            values_diff = _get_values_diff(test_type, series, index, ind)
            temp_integer_index = series_t.index.get_loc(ind)
            s_n_t = 1 / (sigma_sq_t * np.sqrt(integer_index - temp_integer_index)) * values_diff
            if s_n_t > max_s_n_value:
                max_s_n_value = s_n_t
                max_s_n_critical_value = np.sqrt(
                    4.6 + np.log(integer_index - temp_integer_index))  # 4.6 is b_a estimate derived via Monte-Carlo

        s_n_t_series.loc[index, ['stat', 'critical_value']] = max_s_n_value, max_s_n_critical_value
    return s_n_t_series


def get_chu_stinchcombe_white_statistics(series: pd.Series, test_type: str = 'one_sided',
                                         num_threads: int = 8, verbose: bool = True) -> pd.Series:
    """
    Multithread Chu-Stinchcombe-White test implementation, p.251

    :param series: (pd.Series) Series to get statistics for
    :param test_type: (str): Two-sided or one-sided test
    :param num_threads: (int) Number of cores
    :param verbose: (bool) Flag to report progress on asynch jobs
    :return: (pd.Series) Statistics
    """
    molecule = series.index[2:series.shape[0]]  # For the first two values we don't have enough info

    s_n_t_series = mp_pandas_obj(func=_get_s_n_for_t,
                                 pd_obj=('molecule', molecule),
                                 series=series,
                                 test_type=test_type,
                                 num_threads=num_threads,
                                 verbose=verbose,
                                 )
    return s_n_t_series



"""
This module contains the code snippets found in Chapter 10 of "Advances in Financial Machine Learning" by
Marcos López de Prado. The code has been amended for readability, to conform to PEP8 rules, to keep the snippets as
manageable single-units of functionality, as well as to account for deprecation of functions originally used, but is
otherwise unaltered.
"""

import warnings
import pandas as pd
import numpy as np
from scipy.stats import norm

from famlafl.util.multiprocess import mp_pandas_obj


def get_signal(prob, num_classes, pred=None):
    """
    SNIPPET 10.1 - FROM PROBABILITIES TO BET SIZE
    Calculates the given size of the bet given the side and the probability (i.e. confidence) of the prediction. In this
    representation, the probability will always be between 1/num_classes and 1.0.

    :param prob: (pd.Series) The probability of the predicted bet side.
    :param num_classes: (int) The number of predicted bet sides.
    :param pred: (pd.Series) The predicted bet side. Default value is None which will return a relative bet size
     (i.e. without multiplying by the side).
    :return: (pd.Series) The bet size.
    """
    # Get signals from predictions.
    if prob.shape[0] == 0:
        return pd.Series(dtype='float64')

    # 1) Generate signals from multinomial classification (one-vs-rest).
    bet_sizes = (prob - 1/num_classes) / (prob * (1 - prob))**0.5

    # Allow for bet size to be returned with or without side.
    if not isinstance(pred, type(None)):
        # signal = side * size
        bet_sizes = pred * (2 * norm.cdf(bet_sizes) - 1)
    else:
        # signal = size only
        bet_sizes = bet_sizes.apply(lambda s: 2 * norm.cdf(s) - 1)

    # Note 1: In the book, this function contains a conditional statement checking for a column named 'side',
    # then executes what is essentially the above line. This has been removed as it appears to be redundant
    # and simplifies the function.

    # Note 2: In the book, this function includes the averaging and discretization steps, which are omitted here.
    # The functions for performing these are included in this file, and can be applied as options in the user-level
    # functions in bet_sizing.py.

    return bet_sizes


def avg_active_signals(signals, num_threads=1):
    """
    SNIPPET 10.2 - BETS ARE AVERAGED AS LONG AS THEY ARE STILL ACTIVE
    Function averages the bet sizes of all concurrently active bets. This function makes use of multiprocessing.

    :param signals: (pandas.DataFrame) Contains at least the following columns:
     'signal' - the bet size
     't1' - the closing time of the bet
     And the index must be datetime format.
    :param num_threads: (int) Number of threads to use in multiprocessing, default value is 1.
    :return: (pandas.Series) The averaged bet sizes.
    """
    if signals.empty:
        return pd.Series(dtype='float64')
        
    # 1) Time points where signals change (either one start or one ends).
    t_pnts = set(signals['t1'].dropna().to_numpy())
    t_pnts = t_pnts.union(signals.index.to_numpy())
    t_pnts = list(t_pnts)
    t_pnts.sort()
    
    if len(t_pnts) == 0:
        return pd.Series(dtype='float64')
        
    # If only one thread, process directly without multiprocessing
    if num_threads == 1:
        return mp_avg_active_signals(signals, t_pnts)
        
    # Use multiprocessing for multiple threads
    out = mp_pandas_obj(mp_avg_active_signals, ('molecule', t_pnts), num_threads, signals=signals)
    return out


def mp_avg_active_signals(signals, molecule, **kwargs):
    """
    Part of SNIPPET 10.2
    A function to be passed to the 'mp_pandas_obj' function to allow the bet sizes to be averaged using multiprocessing.

    At time loc, average signal among those still active.
    Signal is active if (a) it is issued before or at loc, and (b) loc is before the signal's end time,
    or end time is still unknown (NaT).

    :param signals: (pandas.DataFrame) Contains at least the following columns: 'signal' (the bet size) and 't1' (the closing time of the bet).
    :param molecule: (list) Indivisible tasks to be passed to 'mp_pandas_obj', in this case a list of datetimes.
    :param kwargs: Additional keyword arguments (ignored)
    :return: (pandas.Series) The averaged bet size sub-series.
    """
    if signals.empty or len(molecule) == 0:
        return pd.Series(dtype='float64')
        
    out = pd.Series(dtype='float64')
    for loc in molecule:
        df0 = (signals.index.to_numpy() <= loc)&((loc < signals['t1'])|pd.isnull(signals['t1']))
        act = signals[df0].index
        if act.size > 0:
            # Average active signals if they exist.
            out[loc] = signals.loc[act, 'signal'].mean()
        else:
            # Return zero if no signals are active at this time step.
            out[loc] = 0

    return out


def discrete_signal(signal0, step_size):
    """
    SNIPPET 10.3 - SIZE DISCRETIZATION TO PREVENT OVERTRADING
    Discretizes the bet size signal based on the step size given.

    :param signal0: (pandas.Series) The signal to discretize.
    :param step_size: (float) Step size.
    :return: (pandas.Series) The discretized signal.
    """
    signal1 = (signal0 / step_size).round() * step_size
    signal1[signal1 > 1] = 1  # Cap
    signal1[signal1 < -1] = -1  # Floor
    return signal1


# ==============================================================================
# SNIPPET 10.4 - DYNAMIC POSITION SIZE AND LIMIT PRICE
# The below functions are part of or derived from the functions
# in snippet 10.4.
# ==============================================================================
# Bet size calculations based on a sigmoid function.
def bet_size_sigmoid(w_param, price_div):
    """
    Part of SNIPPET 10.4
    Calculates the bet size from the price divergence and a regulating coefficient.
    Based on a sigmoid function for a bet size algorithm.

    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param price_div: (float) Price divergence, forecast price - market price.
    :return: (float) The bet size.
    """
    return price_div * ((w_param + price_div**2)**(-0.5))


def get_target_pos_sigmoid(w_param, forecast_price, market_price, max_pos):
    """
    Part of SNIPPET 10.4
    Calculates the target position given the forecast price, market price, maximum position size, and a regulating
    coefficient. Based on a sigmoid function for a bet size algorithm.

    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param forecast_price: (float) Forecast price.
    :param market_price: (float) Market price.
    :param max_pos: (int) Maximum absolute position size.
    :return: (int) Target position.
    """
    return int(bet_size_sigmoid(w_param, forecast_price-market_price) * max_pos)


def inv_price_sigmoid(forecast_price, w_param, m_bet_size):
    """
    Part of SNIPPET 10.4
    Calculates the inverse of the bet size with respect to the market price.
    Based on a sigmoid function for a bet size algorithm.

    :param forecast_price: (float) Forecast price.
    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param m_bet_size: (float) Bet size.
    :return: (float) Inverse of bet size with respect to market price.
    """
    return forecast_price - m_bet_size * (w_param / (1 - m_bet_size**2))**0.5


def limit_price_sigmoid(target_pos, pos, forecast_price, w_param, max_pos):
    """
    Part of SNIPPET 10.4
    Calculates the limit price.
    Based on a sigmoid function for a bet size algorithm.

    :param target_pos: (int) Target position.
    :param pos: (int) Current position.
    :param forecast_price: (float) Forecast price.
    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param max_pos: (int) Maximum absolute position size.
    :return: (float) Limit price.
    """
    if target_pos == pos:
        # Return NaN if the current and target positions are the same to avoid divide-by-zero error.
        return np.nan

    sgn = np.sign(target_pos-pos)
    l_p = 0
    for j in range(abs(pos+sgn), abs(target_pos+1)):
        l_p += inv_price_sigmoid(forecast_price, w_param, j/float(max_pos))

    l_p = l_p / abs(target_pos-pos)
    return l_p


def get_w_sigmoid(price_div, m_bet_size):
    """
    Part of SNIPPET 10.4
    Calculates the inverse of the bet size with respect to the regulating coefficient 'w'.
    Based on a sigmoid function for a bet size algorithm.

    :param price_div: (float) Price divergence, forecast price - market price.
    :param m_bet_size: (float) Bet size.
    :return: (float) Inverse of bet size with respect to the
        regulating coefficient.
    """
    return (price_div**2) * ((m_bet_size**(-2)) - 1)


# ==============================================================================
# Bet size calculations based on a power function.
def bet_size_power(w_param, price_div):
    """
    Derived from SNIPPET 10.4
    Calculates the bet size from the price divergence and a regulating coefficient.
    Based on a power function for a bet size algorithm.

    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param price_div: (float) Price divergence, f - market_price, must be between -1 and 1, inclusive.
    :return: (float) The bet size.
    """
    if not (-1 <= price_div <= 1):
        raise ValueError(f"Price divergence must be between -1 and 1, inclusive. Found price divergence value:"
                         f" {price_div}")
    if price_div == 0.0:
        return 0.0

    return np.sign(price_div) * abs(price_div)**w_param


def get_target_pos_power(w_param, forecast_price, market_price, max_pos):
    """
    Derived from SNIPPET 10.4
    Calculates the target position given the forecast price, market price, maximum position size, and a regulating
    coefficient. Based on a power function for a bet size algorithm.

    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param forecast_price: (float) Forecast price.
    :param market_price: (float) Market price.
    :param max_pos: (float) Maximum absolute position size.
    :return: (float) Target position.
    """
    return int(bet_size_power(w_param, forecast_price-market_price) * max_pos)


def inv_price_power(forecast_price, w_param, m_bet_size):
    """
    Derived from SNIPPET 10.4
    Calculates the inverse of the bet size with respect to the market price.
    Based on a power function for a bet size algorithm.

    :param forecast_price: (float) Forecast price.
    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param m_bet_size: (float) Bet size.
    :return: (float) Inverse of bet size with respect to market price.
    """
    if m_bet_size == 0.0:
        return forecast_price
    return forecast_price - np.sign(m_bet_size) * abs(m_bet_size)**(1/w_param)


def limit_price_power(target_pos, pos, forecast_price, w_param, max_pos):
    """
    Derived from SNIPPET 10.4
    Calculates the limit price. Based on a power function for a bet size algorithm.

    :param target_pos: (float) Target position.
    :param pos: (float) Current position.
    :param forecast_price: (float) Forecast price.
    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param max_pos: (float) Maximum absolute position size.
    :return: (float) Limit price.
    """
    sgn = np.sign(target_pos-pos)
    l_p = 0
    for j in range(abs(pos+sgn), abs(target_pos+1)):
        l_p += inv_price_power(forecast_price, w_param, j/float(max_pos))

    l_p = l_p / abs(target_pos-pos)
    return l_p


def get_w_power(price_div, m_bet_size):
    """
    Derived from SNIPPET 10.4
    Calculates the inverse of the bet size with respect to the regulating coefficient 'w'.
    The 'w' coefficient must be greater than or equal to zero.
    Based on a power function for a bet size algorithm.

    :param price_div: (float) Price divergence, forecast price - market price.
    :param m_bet_size: (float) Bet size.
    :return: (float) Inverse of bet size with respect to the regulating coefficient.
    """
    if not -1 <= price_div <= 1:
        raise ValueError("Price divergence argument 'x' must be between -1 and 1,"
                         " inclusive when using function 'power'.")

    w_calc = np.log(m_bet_size/np.sign(price_div)) / np.log(abs(price_div))
    if w_calc < 0:
        warnings.warn("'w' parameter evaluates to less than zero. Zero is returned.", UserWarning)

    return max(0, w_calc)


# ==============================================================================
# Bet size calculation functions, power and sigmoid packaged together.
# This is useful as more bet sizing function options are added.
def bet_size(w_param, price_div, func):
    """
    Derived from SNIPPET 10.4
    Calculates the bet size from the price divergence and a regulating coefficient.
    The 'func' argument allows the user to choose between bet sizing functions.

    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param price_div: (float) Price divergence, f - market_price
    :param func: (string) Function to use for dynamic calculation. Valid options are: 'sigmoid', 'power'.
    :return: (float) The bet size.
    """
    return {'sigmoid': bet_size_sigmoid,
            'power': bet_size_power}[func](w_param, price_div)


def get_target_pos(w_param, forecast_price, market_price, max_pos, func):
    """
    Derived from SNIPPET 10.4
    Calculates the target position given the forecast price, market price, maximum position size, and a regulating
    coefficient. The 'func' argument allows the user to choose between bet sizing functions.

    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param forecast_price: (float) Forecast price.
    :param market_price: (float) Market price.
    :param max_pos: (int) Maximum absolute position size.
    :param func: (string) Function to use for dynamic calculation. Valid options are: 'sigmoid', 'power'.
    :return: (int) Target position.
    """
    return {'sigmoid': get_target_pos_sigmoid,
            'power': get_target_pos_power}[func](w_param, forecast_price, market_price, max_pos)


def inv_price(forecast_price, w_param, m_bet_size, func):
    """
    Derived from SNIPPET 10.4
    Calculates the inverse of the bet size with respect to the market price.
    The 'func' argument allows the user to choose between bet sizing functions.

    :param forecast_price: (float) Forecast price.
    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param m_bet_size: (float) Bet size.
    :return: (float) Inverse of bet size with respect to market price.
    """
    return {'sigmoid': inv_price_sigmoid,
            'power': inv_price_power}[func](forecast_price, w_param, m_bet_size)


def limit_price(target_pos, pos, forecast_price, w_param, max_pos, func):
    """
    Derived from SNIPPET 10.4
    Calculates the limit price. The 'func' argument allows the user to choose between bet sizing functions.

    :param target_pos: (int) Target position.
    :param pos: (int) Current position.
    :param forecast_price: (float) Forecast price.
    :param w_param: (float) Coefficient regulating the width of the bet size function.
    :param max_pos: (int) Maximum absolute position size.
    :param func: (string) Function to use for dynamic calculation. Valid options are: 'sigmoid', 'power'.
    :return: (float) Limit price.
    """
    return {'sigmoid': limit_price_sigmoid,
            'power': limit_price_power}[func](int(target_pos), int(pos), forecast_price, w_param, max_pos)


def get_w(price_div, m_bet_size, func):
    """
    Derived from SNIPPET 10.4
    Calculates the inverse of the bet size with respect to the regulating coefficient 'w'.
    The 'func' argument allows the user to choose between bet sizing functions.

    :param price_div: (float) Price divergence, forecast price - market price.
    :param m_bet_size: (float) Bet size.
    :param func: (string) Function to use for dynamic calculation. Valid options are: 'sigmoid', 'power'.
    :return: (float) Inverse of bet size with respect to the regulating coefficient.
    """
    return {'sigmoid': get_w_sigmoid,
            'power': get_w_power}[func](price_div, m_bet_size)



"""
Implementation of distance using the Generic Non-Parametric Representation approach from "Some contributions to the
clustering of financial time series and applications to credit default swaps" by Gautier Marti
https://www.researchgate.net/publication/322714557
"""
import numpy as np
from scipy.stats import spearmanr

# pylint: disable=invalid-name

def spearmans_rho(x: np.array, y: np.array) -> float:
    """
    Calculates a statistical estimate of Spearman's rho - a copula-based dependence measure.

    Formula for calculation:
    rho = 1 - (6)/(T*(T^2-1)) * Sum((X_t-Y_t)^2)

    It is more robust to noise and can be defined if the variables have an infinite second moment.
    This statistic is described in more detail in the work by Gautier Marti
    https://www.researchgate.net/publication/322714557 (p.54)

    This method is a wrapper for the scipy spearmanr function. For more details about the function and its parameters,
    please visit scipy documentation
    https://docs.scipy.org/doc/scipy-0.14.0/reference/generated/scipy.stats.spearmanr.html

    :param x: (np.array/pd.Series) X vector
    :param y: (np.array/pd.Series) Y vector (same number of observations as X)
    :return: (float) Spearman's rho statistical estimate
    """

    # Coefficient calculationS
    rho, _ = spearmanr(x, y)

    return rho

def gpr_distance(x: np.array, y: np.array, theta: float) -> float:
    """
    Calculates the distance between two Gaussians under the Generic Parametric Representation (GPR) approach.

    According to the original work https://www.researchgate.net/publication/322714557 (p.70):
    "This is a fast and good proxy for distance d_theta when the first two moments ... predominate". But it's not
    a good metric for heavy-tailed distributions.

    Parameter theta defines what type of information dependency is being tested:
    - for theta = 0 the distribution information is tested
    - for theta = 1 the dependence information is tested
    - for theta = 0.5 a mix of both information types is tested

    With theta in [0, 1] the distance lies in range [0, 1] and is a metric. (See original work for proof, p.71)

    :param x: (np.array/pd.Series) X vector.
    :param y: (np.array/pd.Series) Y vector (same number of observations as X).
    :param theta: (float) Type of information being tested. Falls in range [0, 1].
    :return: (float) Distance under GPR approach.
    """

    # Calculating the GPR distance
    distance = theta * (1 - spearmans_rho(x, y)) / 2 + \
               (1 - theta) * (1 - ((2 * x.std() * y.std()) / (x.std()**2 + y.std()**2))**(1/2) *
                              np.exp(- (1 / 4) * (x.mean() - y.mean())**2 / (x.std()**2 + y.std()**2)))

    return distance**(1/2)

def gnpr_distance(x: np.array, y: np.array, theta: float, bandwidth: float = 0.01) -> float:
    """
    Calculates the empirical distance between two random variables under the Generic Non-Parametric Representation
    (GNPR) approach.

    Formula for the distance is taken from https://www.researchgate.net/publication/322714557 (p.72).

    Parameter theta defines what type of information dependency is being tested:
    - for theta = 0 the distribution information is tested
    - for theta = 1 the dependence information is tested
    - for theta = 0.5 a mix of both information types is tested

    With theta in [0, 1] the distance lies in the range [0, 1] and is a metric. (See original work for proof, p.71)

    :param x: (np.array/pd.Series) X vector.
    :param y: (np.array/pd.Series) Y vector (same number of observations as X).
    :param theta: (float) Type of information being tested. Falls in range [0, 1].
    :param bandwidth: (float) Bandwidth to use for splitting the X and Y vector observations. (0.01 by default)
    :return: (float) Distance under GNPR approach.
    """

    # Number of observations
    num_obs = x.shape[0]

    # Calculating the d_1 distance
    dist_1 = 3 / (num_obs * (num_obs**2 - 1)) * (np.power(x - y, 2).sum())

    # Creating the proper bins
    min_val = min(x.min(), y.min())
    max_val = max(x.max(), y.max())

    # Creating a grid and histograms
    bins = np.arange(min_val, max_val + bandwidth, bandwidth)
    hist_x = np.histogram(x, bins)[0] / num_obs
    hist_y = np.histogram(y, bins)[0] / num_obs

    # Calculating the d_0 distance
    dist_0 = np.power(hist_x**(1/2) - hist_y**(1/2), 2).sum() / 2

    # Calculating the GNPR distance
    distance = theta * dist_1 + (1 - theta) * dist_0

    return distance**(1/2)



"""
Test various functions regarding datasets load.
"""

import unittest
from famlafl.datasets.load_datasets import (load_tick_sample, load_stock_prices, load_dollar_bar_sample)


class TestDatasets(unittest.TestCase):
    """
    Test load_tick_sample, load_stock_prices, load_dollar_bar_sample.
    """

    def test_load_tick_sample(self):
        """
        Test load_tick_sample function.
        """

        tick_sample_df = load_tick_sample()
        self.assertEqual(tick_sample_df.shape[0], 100)
        self.assertTrue('Price' in tick_sample_df.columns)
        self.assertTrue('Volume' in tick_sample_df.columns)

    def test_load_stock_prices(self):
        """
        Test load_stock_prices function.
        """
        stock_prices_df = load_stock_prices()
        self.assertEqual(stock_prices_df.shape[0], 2141)
        for ticker in ['EEM', 'EWG', 'TIP', 'EWJ', 'EFA', 'IEF', 'EWQ', 'EWU', 'XLB', 'XLE', 'XLF', 'LQD', 'XLK', 'XLU',
                       'EPP', 'FXI', 'VGK', 'VPL', 'SPY', 'TLT', 'BND', 'CSJ', 'DIA']:
            self.assertTrue(ticker in stock_prices_df.columns)

    def test_load_dollar_bar_sample(self):
        """
        Test load_dollar_bar_samples.
        """
        dollar_bars_df = load_dollar_bar_sample()
        self.assertEqual(dollar_bars_df.shape[0], 1000)
        for col in ['open', 'high', 'low', 'close', 'cum_vol', 'cum_dollar', 'cum_ticks']:
            self.assertTrue(col in dollar_bars_df.columns)



"""
Tests Passive Aggressive Mean Reversion.
"""
from unittest import TestCase
import os
import numpy as np
import pandas as pd
from famlafl.online_portfolio_selection.pamr import \
    PAMR


class TestPassiveAggressiveMeanReversion(TestCase):
    # pylint: disable=too-many-public-methods
    # pylint: disable=unsubscriptable-object
    """
    Tests different functions of the Passive Aggressive Mean Reversion class.
    """

    def setUp(self):
        """
        Set the file path for the tick data csv.
        """
        # Set project path to current directory.
        project_path = os.path.dirname(__file__)
        # Add new data path to match stock_prices.csv data.
        data_path = project_path + '/test_data/stock_prices.csv'
        # Read csv, parse dates, and drop NaN.
        self.data = pd.read_csv(data_path, parse_dates=True, index_col="Date").dropna(axis=1)

    def test_pamr_solution(self):
        """
        Test the calculation of passive aggressive mean reversion with the original optimization
        method.
        """
        # Initialize PAMR.
        pamr = PAMR(optimization_method=0, epsilon=0.5, agg=10)
        # Allocates asset prices to PAMR.
        pamr.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(pamr.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_pamr1_solution(self):
        """
        Test the calculation of passive aggressive mean reversion with PAMR-1 optimization method.
        """
        # Initialize PAMR-1.
        pamr1 = PAMR(optimization_method=1, epsilon=0.5, agg=10)
        # Allocates asset prices to PAMR.
        pamr1.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(pamr1.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_pamr2_solution(self):
        """
        Test the calculation of passive aggressive mean reversion with the PAMR-2 optimization method
        """
        # Initialize PAMR-2.
        pamr2 = PAMR(optimization_method=2, epsilon=0.5, agg=10)
        # Allocates asset prices to PAMR.
        pamr2.allocate(self.data, resample_by='M')
        # Create np.array of all_weights.
        all_weights = np.array(pamr2.all_weights)
        # Check if all weights sum to 1.
        for i in range(all_weights.shape[0]):
            weights = all_weights[i]
            assert (weights >= 0).all()
            assert len(weights) == self.data.shape[1]
            np.testing.assert_almost_equal(np.sum(weights), 1)

    def test_pamr_epsilon_error(self):
        """
        Tests ValueError if epsilon is less than 0.
        """
        # Initialize PAMR.
        pamr3 = PAMR(optimization_method=2, epsilon=-1, agg=10)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            pamr3.allocate(self.data)

    def test_pamr_agg_error(self):
        """
        Tests ValueError if aggressiveness is less than 0.
        """
        # Initialize PAMR.
        pamr4 = PAMR(optimization_method=2, epsilon=0.5, agg=-5)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            pamr4.allocate(self.data)

    def test_pamr_method_error(self):
        """
        Tests ValueError if optimization method is not 0, 1, or 2.
        """
        # Initialize PAMR.
        pamr5 = PAMR(optimization_method=5, epsilon=0.5, agg=10)
        with self.assertRaises(ValueError):
            # Running allocate will raise ValueError.
            pamr5.allocate(self.data)



"""
Logic regarding concurrent labels from chapter 4.
"""

import pandas as pd

from famlafl.util.multiprocess import mp_pandas_obj


def num_concurrent_events(close_series_index, label_endtime, molecule):
    """
    Advances in Financial Machine Learning, Snippet 4.1, page 60.

    Estimating the Uniqueness of a Label

    This function uses close series prices and label endtime (when the first barrier is touched) to compute the number
    of concurrent events per bar.

    :param close_series_index: (pd.Series) Close prices index
    :param label_endtime: (pd.Series) Label endtime series (t1 for triple barrier events)
    :param molecule: (an array) A set of datetime index values for processing
    :return: (pd.Series) Number concurrent labels for each datetime index
    """
    # Find events that span the period [molecule[0], molecule[1]]
    label_endtime = label_endtime.fillna(close_series_index[-1])  # Unclosed events still must impact other weights
    label_endtime = label_endtime[label_endtime >= molecule[0]]  # Events that end at or after molecule[0]
    # Events that start at or before t1[molecule].max()
    label_endtime = label_endtime.loc[:label_endtime[molecule].max()]

    # Count events spanning a bar
    nearest_index = close_series_index.searchsorted(pd.DatetimeIndex([label_endtime.index[0], label_endtime.max()]))
    count = pd.Series(0, index=close_series_index[nearest_index[0]:nearest_index[1] + 1])
    for t_in, t_out in label_endtime.items():
        count.loc[t_in:t_out] += 1
    return count.loc[molecule[0]:label_endtime[molecule].max()]


def _get_average_uniqueness(label_endtime, num_conc_events, molecule):
    """
    Advances in Financial Machine Learning, Snippet 4.2, page 62.

    Estimating the Average Uniqueness of a Label

    This function uses close series prices and label endtime (when the first barrier is touched) to compute the number
    of concurrent events per bar.

    :param label_endtime: (pd.Series) Label endtime series (t1 for triple barrier events)
    :param num_conc_events: (pd.Series) Number of concurrent labels (output from num_concurrent_events function).
    :param molecule: (an array) A set of datetime index values for processing.
    :return: (pd.Series) Average uniqueness over event's lifespan.
    """
    # Derive average uniqueness over the event's lifespan
    wght = pd.Series(index=molecule, dtype='float64')
    for t_in, t_out in label_endtime.loc[wght.index].items():
        wght.loc[t_in] = (1. / num_conc_events.loc[t_in:t_out]).mean()
    return wght


def get_av_uniqueness_from_triple_barrier(triple_barrier_events, close_series, num_threads, verbose=True):
    """
    This function is the orchestrator to derive average sample uniqueness from a dataset labeled by the triple barrier
    method.

    :param triple_barrier_events: (pd.DataFrame) Events from labeling.get_events()
    :param close_series: (pd.Series) Close prices.
    :param num_threads: (int) The number of threads concurrently used by the function.
    :param verbose: (bool) Flag to report progress on asynch jobs
    :return: (pd.Series) Average uniqueness over event's lifespan for each index in triple_barrier_events
    """
    out = pd.DataFrame()
    num_conc_events = mp_pandas_obj(num_concurrent_events, ('molecule', triple_barrier_events.index), num_threads,
                                    close_series_index=close_series.index, label_endtime=triple_barrier_events['t1'],
                                    verbose=verbose)
    num_conc_events = num_conc_events.loc[~num_conc_events.index.duplicated(keep='last')]
    num_conc_events = num_conc_events.reindex(close_series.index).fillna(0)
    out['tW'] = mp_pandas_obj(_get_average_uniqueness, ('molecule', triple_barrier_events.index), num_threads,
                              label_endtime=triple_barrier_events['t1'], num_conc_events=num_conc_events,
                              verbose=verbose)
    return out
